# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function, unicode_literals

import logging
import os
import shutil
from operator import itemgetter

import geojson
from pkg_resources import resource_filename
from tqdm import tqdm
from whoosh import fields
from whoosh.fields import Schema
from whoosh.index import create_in, exists_in, open_dir
from whoosh.qparser import QueryParser

from eodag.api.search_result import SearchResult
from eodag.config import (
    SimpleYamlProxyConfig,
    load_default_config,
    override_config_from_env,
    override_config_from_file,
)
from eodag.plugins.manager import PluginManager
from eodag.utils import makedirs
from eodag.utils.exceptions import (
    NoMatchingProductType,
    PluginImplementationError,
    UnsupportedProvider,
)

logger = logging.getLogger("eodag.core")

# pagination defaults
DEFAULT_PAGE = 1
DEFAULT_ITEMS_PER_PAGE = 20


class EODataAccessGateway(object):
    """An API for downloading a wide variety of geospatial products originating
    from different types of providers.

    :param user_conf_file_path: Path to the user configuration file
    :type user_conf_file_path: str or unicode
    """

    def __init__(self, user_conf_file_path=None):
        self.product_types_config = SimpleYamlProxyConfig(
            resource_filename("eodag", os.path.join("resources/", "product_types.yml"))
        )
        self.providers_config = load_default_config()

        self.conf_dir = os.path.join(os.path.expanduser("~"), ".config", "eodag")
        makedirs(self.conf_dir)

        # First level override: From a user configuration file
        if user_conf_file_path is None:
            env_var_name = "EODAG_CFG_FILE"
            standard_configuration_path = os.path.join(self.conf_dir, "eodag.yml")
            user_conf_file_path = os.getenv(env_var_name)
            if user_conf_file_path is None:
                user_conf_file_path = standard_configuration_path
                if not os.path.isfile(standard_configuration_path):
                    shutil.copy(
                        resource_filename(
                            "eodag", os.path.join("resources", "user_conf_template.yml")
                        ),
                        standard_configuration_path,
                    )
        override_config_from_file(self.providers_config, user_conf_file_path)

        # Second level override: From environment variables
        override_config_from_env(self.providers_config)

        self._plugins_manager = PluginManager(self.providers_config)

        # Build a search index for product types
        self._product_types_index = None
        self.build_index()

    def build_index(self):
        """Build a `Whoosh <https://whoosh.readthedocs.io/en/latest/index.html>`_
        index for product types searches.

        .. versionadded:: 1.0
        """
        index_dir = os.path.join(self.conf_dir, ".index")

        # Handle Python 2/3 compatibility: if index_dir exists and contains a whoosh
        # index, it means it was potentially created with a previous Python 3 eodag
        # session, thus using pickle highest protocol version (version 4 at the time of
        # writing for Python 3). In that case, if the current eodag session is in Python
        # 2 a ValueError will be raised saying that the pickle protocol used to
        # serialized the previous index is incompatible with the current Python version.
        # If that's the case, we first remove the problematic index and create another
        # one from scratch
        try:
            create_index = not exists_in(index_dir)
        except ValueError as ve:
            if "unsupported pickle protocol" in ve.message:
                shutil.rmtree(index_dir)
                create_index = True
            else:
                raise

        if create_index:
            logger.debug("Creating product types index in %s", index_dir)
            makedirs(index_dir)
            product_types_schema = Schema(
                ID=fields.STORED,
                abstract=fields.TEXT,
                instrument=fields.IDLIST,
                platform=fields.ID,
                platformSerialIdentifier=fields.IDLIST,
                processingLevel=fields.ID,
                sensorType=fields.ID,
            )
            self._product_types_index = create_in(index_dir, product_types_schema)
            ix_writer = self._product_types_index.writer()
            for product_type in self.list_product_types():
                ix_writer.add_document(**product_type)
            ix_writer.commit()
        else:
            if self._product_types_index is None:
                logger.debug("Opening product types index in %s", index_dir)
                self._product_types_index = open_dir(index_dir)

    def set_preferred_provider(self, provider):
        """Set max priority for the given provider.

        >>> import tempfile, os
        >>> config = tempfile.NamedTemporaryFile(delete=True)
        >>> dag = EODataAccessGateway(user_conf_file_path=os.path.join(
        ...     tempfile.gettempdir(), config.name))
        >>> # This also tests get_preferred_provider method by the way
        >>> dag.get_preferred_provider()
        ('peps', 1)
        >>> # For the following lines, see
        >>> # http://python3porting.com/problems.html#handling-expected-exceptions
        >>> import eodag.utils.exceptions
        >>> try:
        ...     dag.set_preferred_provider(u'unknown')
        ...     raise AssertionError(u'UnsupportedProvider exception was not raised'
        ...                           'as expected')
        ... except eodag.utils.exceptions.UnsupportedProvider:
        ...     pass
        >>> dag.set_preferred_provider(u'usgs')
        >>> dag.get_preferred_provider()
        ('usgs', 2)
        >>> dag.set_preferred_provider(u'theia')
        >>> dag.get_preferred_provider()
        ('theia', 3)
        >>> dag.set_preferred_provider(u'usgs')
        >>> dag.get_preferred_provider()
        ('usgs', 4)
        >>> config.close()

        :param provider: The name of the provider that should be considered as the
                         preferred provider to be used for this instance
        :type provider: str or unicode
        """
        if provider not in self.available_providers():
            raise UnsupportedProvider("This provider is not recognised by eodag")
        preferred_provider, max_priority = self.get_preferred_provider()
        if preferred_provider != provider:
            new_priority = max_priority + 1
            self._plugins_manager.set_priority(provider, new_priority)

    def get_preferred_provider(self):
        """Get the provider currently set as the preferred one for searching
        products, along with its priority.

        :return: The provider with the maximum priority and its priority
        :rtype: tuple(str, int)
        """
        providers_with_priority = [
            (provider, conf.priority)
            for provider, conf in self.providers_config.items()
        ]
        preferred, priority = max(providers_with_priority, key=itemgetter(1))
        return preferred, priority

    def list_product_types(self, provider=None):
        """Lists supported product types.

        :param provider: The name of a provider that must support the product
                         types we are about to list
        :type provider: str or unicode
        :returns: The list of the product types that can be accessed using eodag.
        :rtype: list(dict)
        :raises: :class:`~eodag.utils.exceptions.UnsupportedProvider`
        """
        product_types = []
        if provider is not None:
            if provider in self.providers_config:
                provider_supported_products = self.providers_config[provider].products
                for product_type_id in provider_supported_products:
                    product_type = dict(
                        ID=product_type_id, **self.product_types_config[product_type_id]
                    )
                    if product_type_id not in product_types:
                        product_types.append(product_type)
                return sorted(product_types, key=itemgetter("ID"))
            raise UnsupportedProvider("The requested provider is not (yet) supported")
        # Only get the product types supported by the available providers
        for provider in self.available_providers():
            current_product_type_ids = [pt["ID"] for pt in product_types]
            product_types.extend(
                [
                    pt
                    for pt in self.list_product_types(provider=provider)
                    if pt["ID"] not in current_product_type_ids
                ]
            )
        # Return the product_types sorted in lexicographic order of their ID
        return sorted(product_types, key=itemgetter("ID"))

    def available_providers(self):
        """Gives the list of the available providers"""
        return sorted(tuple(self.providers_config.keys()))

    def guess_product_type(self, **kwargs):
        """Find the eodag product type code that best matches a set of search params

        >>> from eodag import EODataAccessGateway
        >>> dag = EODataAccessGateway()
        >>> dag.guess_product_type(
        ...     instrument="MSI",
        ...     platform="SENTINEL2",
        ...     platformSerialIdentifier="S2A",
        ... )
        ['S2_MSI_L1C', 'S2_MSI_L2A']
        >>> import eodag.utils.exceptions
        >>> try:
        ...     dag.guess_product_type()
        ...     raise AssertionError(u"NoMatchingProductType exception not raised")
        ... except eodag.utils.exceptions.NoMatchingProductType:
        ...     pass

        :param kwargs: A set of search parameters as keywords arguments
        :return: The best match for the given parameters
        :rtype: str or unicode
        :raises: :class:`~eodag.utils.exceptions.NoMatchingProductType`

        .. versionadded:: 1.0
        """
        supported_params = {
            param
            for param in (
                "instrument",
                "platform",
                "platformSerialIdentifier",
                "processingLevel",
                "sensorType",
            )
            if kwargs.get(param, None) is not None
        }
        with self._product_types_index.searcher() as searcher:
            results = None
            # For each search key, do a guess and then upgrade the result (i.e. when
            # merging results, if a hit appears in both results, its position is raised
            # to the top. This way, the top most result will be the hit that best
            # matches the given queries. Put another way, this best guess is the one
            # that crosses the highest number of search params from the given queries
            for search_key in supported_params:
                query = QueryParser(search_key, self._product_types_index.schema).parse(
                    kwargs[search_key]
                )
                if results is None:
                    results = searcher.search(query)
                else:
                    results.upgrade_and_extend(searcher.search(query))
            guesses = [r["ID"] for r in results or []]
        if guesses:
            return guesses
        raise NoMatchingProductType()

    def search(
        self,
        page=DEFAULT_PAGE,
        items_per_page=DEFAULT_ITEMS_PER_PAGE,
        raise_errors=False,
        start=None,
        end=None,
        box=None,
        **kwargs
    ):
        """Look for products matching criteria on known providers.

        The default behaviour is to look for products on the provider with the
        highest priority supporting the requested product type. These priorities
        are configurable through user configuration file or individual
        environment variable.

        :param page: The page number to return (default: 1)
        :type page: int
        :param items_per_page: The number of results that must appear in one single
                               page (default: 20)
        :type items_per_page: int
        :param raise_errors:  When an error occurs when searching, if this is set to
                              True, the error is raised (default: False)
        :type raise_errors: bool
        :param start: Start sensing time in iso format
        :type start: str or unicode
        :param end: End sensing time in iso format
        :type end: str or unicode
        :param box: A bounding box delimiting the AOI (as a dict with keys: "lonmin",
                    "latmin", "lonmax", "latmax")
        :type box: dict
        :param dict kwargs: some other criteria that will be used to do the search,
                            may content custom parameter ``custom="foo=1&bar=2"``
        :returns: A collection of EO products matching the criteria and the total
                  number of results found
        :rtype: tuple(:class:`~eodag.api.search_result.SearchResult`, int)

        .. versionchanged::
           1.0

                * The ``product_type`` parameter is no longer mandatory
                * Support new search parameters compliant with OpenSearch
                * Fails if a suitable product type could not be guessed (returns an
                  empty search result) and if the user is not querying a specific
                  product (by providing ``id`` as search parameter)
                * A search by ID is now performed if a product type can not be guessed
                  from the user input, and if the user has provided an ID as a search
                  criteria (in the keyword arguments)
                * It is now possible to pass in the name of the provider on which to
                  perform the request when searching a product by its ID,
                  for performance reasons. In that case, the search with the product ID
                  will be done directly on that provider

        .. versionchanged::
           0.7.1

                * The search now stops at the first provider that supports the
                  requested product type (removal of the partial mechanism)
                * The method now returns a tuple of 2 elements, the first one
                  being the result and the second one being the total number of
                  results satisfying the criteria on the provider. This is useful
                  for applications wrapping eodag and wishing to implement
                  pagination. An example of this kind of application is the
                  embedded eodag HTTP server. Also useful for informational purposes
                  in the CLI: the user is informed about the total number of results
                  available, and can ask for retrieving a given number of these
                  results (See the help message of the CLI for more information).

        .. note::
            The search interfaces, which are implemented as plugins, are required to
            return a list as a result of their processing. This requirement is
            enforced here.
        """
        product_type = kwargs.get("productType", None)
        if product_type is None:
            try:
                guesses = self.guess_product_type(**kwargs)
                # By now, only use the best bet
                product_type = guesses[0]
            except NoMatchingProductType:
                queried_id = kwargs.get("id", None)
                if queried_id is None:
                    logger.error("Unable to satisfy search query: %s", kwargs)
                    logger.error(
                        "No product type could be guessed with provided arguments"
                    )
                else:
                    provider = kwargs.get("provider", None)
                    return self._search_by_id(kwargs["id"], provider=provider)
                return SearchResult([]), 0

        kwargs["productType"] = product_type
        if start is not None:
            kwargs["startTimeFromAscendingNode"] = start
        if end is not None:
            kwargs["completionTimeFromAscendingNode"] = end
        if box is not None:
            kwargs["geometry"] = box

        plugin = next(
            self._plugins_manager.get_search_plugins(product_type=product_type)
        )
        logger.info(
            "Searching product type '%s' on provider: %s", product_type, plugin.provider
        )
        logger.debug("Using plugin class for search: %s", plugin.__class__.__name__)
        auth = self._plugins_manager.get_auth_plugin(plugin.provider)
        return self._do_search(
            plugin,
            auth=auth,
            page=page,
            items_per_page=items_per_page,
            raise_errors=raise_errors,
            **kwargs
        )

    def _search_by_id(self, uid, provider=None):
        """Internal method that enables searching a product by its id.

        Keeps requesting providers until a result matching the id is supplied. The
        search plugins should be developed in the way that enable them to handle the
        support of a search by id by the providers. The providers are requested one by
        one, in the order defined by their priorities. Be aware that because of that,
        the search can be slow, if the priority order is such that the provider that
        contains the requested product has the lowest priority. However, you can always
        speed up a little the search by passing the name of the provider on which to
        perform the search, if this information is available

        :param uid: The uid of the EO product
        :type uid: str (Python 3) or unicode (Python 2)
        :param provider: (optional) The provider on which to search the product.
                         This may be useful for performance reasons when the user
                         knows this product is available on the given provider
        :type provider: str (Python 3) or unicode (Python 2)
        :returns: A search result with one EO product or None at all, and the number
                  of EO products retrieved (0 or 1)
        :rtype: tuple(:class:`~eodag.api.search_result.SearchResult`, int)

        .. versionadded:: 1.0
        """
        for plugin in self._plugins_manager.get_search_plugins(provider=provider):
            logger.info(
                "Searching product with id '%s' on provider: %s", uid, plugin.provider
            )
            logger.debug("Using plugin class for search: %s", plugin.__class__.__name__)
            auth = self._plugins_manager.get_auth_plugin(plugin.provider)
            results, _ = self._do_search(plugin, auth=auth, id=uid)
            if len(results) == 1:
                return results, 1
        return SearchResult([]), 0

    def _do_search(self, search_plugin, **kwargs):
        """Internal method that performs a search on a given provider.

        .. versionadded:: 1.0
        """
        results = SearchResult([])
        total_results = 0
        try:
            res, nb_res = search_plugin.query(**kwargs)

            # Only do the pagination computations when it makes sense. For example,
            # for a search by id, we can reasonably guess that the provider will return
            # At most 1 product, so we don't need such a thing as pagination
            page = kwargs.get("page")
            items_per_page = kwargs.get("items_per_page")
            if page and items_per_page:
                # Take into account the fact that a provider may not return the count of
                # products (in that case, fallback to using the length of the results it
                # returned and the page requested. As an example, check the result of
                # the following request (look for the value of properties.totalResults)
                # https://theia-landsat.cnes.fr/resto/api/collections/Landsat/search.json?
                # maxRecords=1&page=1
                if nb_res == 0:
                    nb_res = len(res) * page

                # Attempt to ensure a little bit more coherence. Some providers return
                # a fuzzy number of total results, meaning that you have to keep
                # requesting it until it has returned everything it has to know exactly
                # how many EO products they have in their stock. In that case, we need
                # to replace the returned number of results with the sum of the number
                # of items that were skipped so far and the length of the currently
                # retrieved items. We know there is an incoherence when the number of
                # skipped items is greater than the total number of items returned by
                # the plugin
                nb_skipped_items = items_per_page * (page - 1)
                nb_current_items = len(res)
                if nb_skipped_items > nb_res:
                    if nb_res != 0:
                        nb_res = nb_skipped_items + nb_current_items
                    # This is for when the returned results is an empty list and the
                    # number of results returned is incoherent with the observations.
                    # In that case, we assume the total number of results is the number
                    # of skipped results. By requesting a lower page than the current
                    # one, a user can iteratively reach the last page of results for
                    # these criteria on the provider.
                    else:
                        nb_res = nb_skipped_items

            if not isinstance(res, list):
                raise PluginImplementationError(
                    "The query function of a Search plugin must return a list of "
                    "results, got {} instead".format(type(res))
                )

            # Filter and attach to each eoproduct in the result the plugin capable of
            # downloading it (this is done to enable the eo_product to download itself
            # doing: eo_product.download()). The filtering is done by keeping only
            # those eo_products that intersects the search extent (if there was no
            # search extent, search_intersection contains the geometry of the
            # eo_product)
            # WARNING: this means an eo_product that has an invalid geometry can still
            # be returned as a search result if there was no search extent (because we
            # will not try to do an intersection)
            for eo_product in res:
                if eo_product.search_intersection is not None:
                    download_plugin = self._plugins_manager.get_download_plugin(
                        eo_product
                    )
                    eo_product.register_downloader(
                        download_plugin, kwargs.get("auth", None)
                    )

            results.extend(res)
            total_results += nb_res
            logger.info(
                "Found %s result(s) on provider '%s'", nb_res, search_plugin.provider
            )
        except Exception:
            logger.info(
                "No result from provider '%s' due to an error during search. Raise "
                "verbosity of log messages for details",
                search_plugin.provider,
            )
            if kwargs.get("raise_errors"):
                # Raise the error, letting the application wrapping eodag know that
                # something went bad. This way it will be able to decide what to do next
                raise
            else:
                logger.exception(
                    "Error while searching on provider %s (ignored):",
                    search_plugin.provider,
                )
        return SearchResult(results), total_results

    def crunch(self, results, **kwargs):
        """Apply the filters given through the keyword arguments to the results

        :param results: The results of a eodag search request
        :type results: :class:`~eodag.api.search_result.SearchResult`
        :return: The result of successively applying all the filters to the results
        :rtype: :class:`~eodag.api.search_result.SearchResult`
        """
        search_criteria = kwargs.pop("search_criteria", {})
        for cruncher_name, cruncher_args in kwargs.items():
            cruncher = self._plugins_manager.get_crunch_plugin(
                cruncher_name, **cruncher_args
            )
            results = results.crunch(cruncher, **search_criteria)
        return results

    @staticmethod
    def sort_by_extent(searches):
        """Combines multiple SearchResults and return a list of SearchResults sorted
        by extent.

        :param searches: List of eodag SearchResult
        :type searches: list
        :return: list of :class:`~eodag.api.search_result.SearchResult`
        """
        products_grouped_by_extent = {}

        for search in searches:
            for product in search:
                same_geom = products_grouped_by_extent.setdefault(
                    "".join([str(round(p, 2)) for p in product.geometry.bounds]), []
                )
                same_geom.append(product)

        return [
            SearchResult(products_grouped_by_extent[extent_as_wkb_hex])
            for extent_as_wkb_hex in products_grouped_by_extent
        ]

    def download_all(self, search_result, progress_callback=None):
        """Download all products resulting from a search.

        :param search_result: A collection of EO products resulting from a search
        :type search_result: :class:`~eodag.api.search_result.SearchResult`
        :param progress_callback: (optional) A method or a callable object
                                  which takes a current size and a maximum
                                  size as inputs and handle progress bar
                                  creation and update to give the user a
                                  feedback on the download progress
        :type progress_callback: :class:`~eodag.utils.ProgressCallback` or None
        :returns: A collection of the absolute paths to the downloaded products
        :rtype: list
        """
        paths = []
        if search_result:
            logger.info("Downloading %s products", len(search_result))
            with tqdm(
                search_result, unit="product", desc="Downloading products"
            ) as bar:
                for product in bar:
                    try:
                        paths.append(
                            self.download(product, progress_callback=progress_callback)
                        )
                    except Exception:
                        import traceback as tb

                        logger.warning(
                            "A problem occurred during download of product: %s. "
                            "Skipping it",
                            product,
                        )
                        logger.debug("\n%s", tb.format_exc())
        else:
            logger.info("Empty search result, nothing to be downloaded !")
        return paths

    @staticmethod
    def serialize(search_result, filename="search_results.geojson"):
        """Registers results of a search into a geojson file.

        :param search_result: A collection of EO products resulting from a search
        :type search_result: :class:`~eodag.api.search_result.SearchResult`
        :param filename: The name of the file to generate
        :type filename: str or unicode
        :returns: The name of the created file
        :rtype: str or unicode
        """
        with open(filename, "w") as fh:
            geojson.dump(search_result, fh)
        return filename

    @staticmethod
    def deserialize(filename):
        """Loads results of a search from a geojson file.

        :param filename: A filename containing a search result encoded as a geojson
        :type filename: str or unicode
        :returns: The search results encoded in `filename`
        :rtype: :class:`~eodag.api.search_result.SearchResult`
        """
        with open(filename, "r") as fh:
            return SearchResult.from_geojson(geojson.load(fh))

    def download(self, product, progress_callback=None):
        """Download a single product.

        This is an alias to the method of the same name on
        :class:`~eodag.api.product.EOProduct`, but it performs some additional
        checks like verifying that a downloader and authenticator are registered
        for the product before trying to download it. If the metadata mapping for
        `downloadLink` is set to something that can be interpreted as a link on a
        local filesystem, the download is skipped (by now, only a link starting
        with `file://` is supported). Therefore, any user that knows how to extract
        product location from product metadata on a provider can override the
        `downloadLink` metadata mapping in the right way. For example, using the
        environment variable:
        EODAG__SOBLOO__SEARCH__METADATA_MAPPING__DOWNLOADLINK="file:///{id}" will
        lead to all `EOProduct`s originating from the provider `sobloo` to have their
        `downloadLink` metadata point to something like: "file:///12345-678",
        making this method immediately return the later string without trying
        to download the product.

        :param product: The EO product to download
        :type product: :class:`~eodag.api.product.EOProduct`
        :param progress_callback: (optional) A method or a callable object
                                  which takes a current size and a maximum
                                  size as inputs and handle progress bar
                                  creation and update to give the user a
                                  feedback on the download progress
        :type progress_callback: :class:`~eodag.utils.ProgressCallback` or None
        :returns: The absolute path to the downloaded product in the local filesystem
        :rtype: str or unicode
        :raises: :class:`~eodag.utils.exceptions.PluginImplementationError`
        :raises: :class:`RuntimeError`
        """
        if product.location.startswith("file"):
            logger.info("Local product detected. Download skipped")
            return product.location
        if product.downloader is None:
            auth = product.downloader_auth
            if auth is None:
                auth = self._plugins_manager.get_auth_plugin(product.provider)
            product.register_downloader(
                self._plugins_manager.get_download_plugin(product), auth
            )
        return product.download(progress_callback=progress_callback)

    def get_cruncher(self, name, **options):
        """Build a crunch plugin from a configuration

        :param name: The name of the cruncher to build
        :type name: str of unicode
        :param options: The configuration options of the cruncher
        :type options: dict
        :return: The cruncher named ``name``
        :rtype: :class:`~eodag.plugins.crunch.Crunch`
        """
        plugin_conf = {"name": name}
        plugin_conf.update({key.replace("-", "_"): val for key, val in options.items()})
        return self._plugins_manager.get_crunch_plugin(name, **plugin_conf)



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import importlib
import pkgutil
from contextlib import contextmanager
from functools import partial


def import_all_modules(base_package, depth=1, exclude=()):
    """Import all modules in base_package, including modules in the sub-packages up to `depth` and excluding modules in
    `exclude`.

    Example::

        # base_package
        #    __init__.py
        #    module1.py
        #    subpackage
        #        __init__.py
        #        subsubpackage
        #            __init__.py
        #            module3.py
        #        module2.py
        #        excluded.py

        # Import 'module1' and 'module2'
        import_all_modules(base_package, depth=2, exclude=['excluded'])
        # Import 'module1'
        import_all_modules(base_package)

    :param base_package: The package from where we must import all the modules
    :type base_package: `module`
    :param depth: (optional) If `base_package` has sub packages, import all the modules recursively up to this level.
                  Defaults to 1 (limits to the level of `base_package`)
    :type depth: int
    :param exclude: (optional) The sub packages and modules to ignore while importing. Empty by default
    :type exclude: tuple(str, ...)

    .. note::
        if `package` and `subpackage` have a module of the same name and this name is included in the exclude
        parameter, the module will be excluded from import for both packages. This is intentionally left as is by
        now for simplification
    """
    if depth > 1:
        for _, name, ispkg in pkgutil.iter_modules(base_package.__path__):
            if not exclude or name not in exclude:
                if ispkg:
                    pkg = importlib.import_module('.{}'.format(name), package=base_package.__name__)
                    import_all_modules(pkg, depth=depth - 1, exclude=exclude)
                else:
                    importlib.import_module('.{}'.format(name), package=base_package.__name__)
    for _, name, ispkg in pkgutil.iter_modules(base_package.__path__):
        if (not exclude or name not in exclude) and not ispkg:
            importlib.import_module('.{}'.format(name), package=base_package.__name__)


@contextmanager
def patch_owslib_requests(verify=True):
    """Overrides call to the :func:`requests.request` and :func:`requests.post` functions by
    :func:`owslib.util.openURL` and :func:`owslib.util.http_post` functions, providing some control over how to use
    these functions in `owslib <https://geopython.github.io/OWSLib/>`_.

    :param verify: Whether to verify the use of https or not
    :type verify: bool
    """
    from owslib.util import requests
    old_request = requests.request
    old_post = requests.post
    try:
        requests.request = partial(requests.request, verify=verify)
        requests.post = partial(requests.post, verify=verify)
        yield
    finally:
        requests.request = old_request
        requests.post = old_post



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

import requests

from eodag.plugins.authentication import Authentication
from eodag.plugins.authentication.openid_connect import CodeAuthorizedAuth
from eodag.utils.exceptions import AuthenticationError


class KeycloakOIDCPasswordAuth(Authentication):

    GRANT_TYPE = 'password'
    TOKEN_URL_TEMPLATE = '{auth_base_uri}/realms/{realm}/protocol/openid-connect/token'

    def __init__(self, provider, config):
        super(KeycloakOIDCPasswordAuth, self).__init__(provider, config)
        self.session = requests.Session()

    def authenticate(self):
        response = self.session.post(
            self.TOKEN_URL_TEMPLATE.format(
                auth_base_uri=self.config.auth_base_uri.rstrip('/'),
                realm=self.config.realm
            ),
            data={
                'client_id': self.config.client_id,
                'client_secret': self.config.client_secret,
                'username': self.config.credentials['username'],
                'password': self.config.credentials['password'],
                'grant_type': self.GRANT_TYPE,
            }
        )
        try:
            response.raise_for_status()
        except requests.HTTPError:
            import traceback as tb
            raise AuthenticationError(
                'Something went wrong while trying to get access token:\n{}'.format(tb.format_exc())
            )
        return CodeAuthorizedAuth(
            response.json()['access_token'],
            self.config.token_provision,
            key=getattr(self.config, 'token_qs_key', None)
        )



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function, unicode_literals

import logging
import os
import tempfile

import yaml
import yaml.constructor
import yaml.parser
from pkg_resources import resource_filename

from eodag.utils import merge_mappings, slugify, utf8_everywhere
from eodag.utils.exceptions import ValidationError

logger = logging.getLogger("eodag.config")


class SimpleYamlProxyConfig(object):
    """A simple configuration class acting as a proxy to an underlying dict object
    as returned by yaml.load"""

    def __init__(self, conf_file_path):
        with open(os.path.abspath(os.path.realpath(conf_file_path)), "r") as fh:
            try:
                self.source = yaml.load(fh, Loader=yaml.SafeLoader)
                utf8_everywhere(self.source)
            except yaml.parser.ParserError as e:
                print("Unable to load user configuration file")
                raise e

    def __getitem__(self, item):
        return self.source[item]

    def __contains__(self, item):
        return item in self.source

    def __iter__(self):
        return iter(self.source)

    def items(self):
        """Iterate over keys and values of source"""
        return self.source.items()

    def values(self):
        """Iterate over values of source"""
        return self.source.values()

    def update(self, other):
        """Update a :class:`~eodag.config.SimpleYamlProxyConfig`"""
        if not isinstance(other, self.__class__):
            raise ValueError("'{}' must be of type {}".format(other, self.__class__))
        self.source.update(other.source)


class ProviderConfig(yaml.YAMLObject):
    """Representation of eodag configuration.

    :param name: The name of the provider
    :type name: str or unicode
    :param priority: (optional) The priority of the provider while searching a product.
                     Lower value means lower priority. (Default: 0)
    :type priority: int
    :param api: (optional) The configuration of a plugin of type Api
    :type api: :class:`~eodag.config.PluginConfig`
    :param search: (optional) The configuration of a plugin of type Search
    :type search: :class:`~eodag.config.PluginConfig`
    :param dict products: (optional) The products types supported by the provider
    :param download: (optional) The configuration of a plugin of type Download
    :type download: :class:`~eodag.config.PluginConfig`
    :param auth: (optional) The configuration of a plugin of type Authentication
    :type auth: :class:`~eodag.config.PluginConfig`
    :param dict kwargs: Additional configuration variables for this provider
    """

    yaml_loader = yaml.Loader
    yaml_dumper = yaml.SafeDumper
    yaml_tag = "!provider"

    @classmethod
    def from_yaml(cls, loader, node):
        """Build a :class:`~eodag.config.ProviderConfig` from Yaml"""
        cls.validate(tuple(node_key.value for node_key, _ in node.value))
        for node_key, node_value in node.value:
            if node_key.value == "name":
                node_value.value = slugify(node_value.value).replace("-", "_")
                break
        return loader.construct_yaml_object(node, cls)

    @classmethod
    def from_mapping(cls, mapping):
        """Build a :class:`~eodag.config.ProviderConfig` from a mapping"""
        cls.validate(mapping)
        for key in ("api", "search", "download", "auth"):
            if key in mapping:
                mapping[key] = PluginConfig.from_mapping(mapping[key])
        c = cls()
        c.__dict__.update(mapping)
        return c

    @staticmethod
    def validate(config_keys):
        """Validate a :class:`~eodag.config.ProviderConfig`"""
        if "name" not in config_keys:
            raise ValidationError("Provider config must have name key")
        if not any(k in config_keys for k in ("api", "search", "download", "auth")):
            raise ValidationError("A provider must implement at least one plugin")
        if "api" in config_keys and any(
            k in config_keys for k in ("search", "download", "auth")
        ):
            raise ValidationError(
                "A provider implementing an Api plugin must not implement any other "
                "type of plugin"
            )

    def update(self, mapping):
        """Update the configuration parameters with values from `mapping`

        :param dict mapping: The mapping from which to override configuration parameters
        """
        if mapping is None:
            mapping = {}
        merge_mappings(
            self.__dict__,
            {
                key: value
                for key, value in mapping.items()
                if key not in ("name", "api", "search", "download", "auth")
                and value is not None
            },
        )
        for key in ("api", "search", "download", "auth"):
            current_value = getattr(self, key, None)
            if current_value is not None:
                current_value.update(mapping.get(key, {}))


class PluginConfig(yaml.YAMLObject):
    """Representation of a plugin config

    :param name: The name of the plugin class to use to instantiate the plugin object
    :type name: str or unicode
    :param dict metadata_mapping: (optional) The mapping between eodag metadata and
                                  the plugin specific metadata
    :param dict free_params: (optional) Additional configuration parameters
    """

    yaml_loader = yaml.Loader
    yaml_dumper = yaml.SafeDumper
    yaml_tag = "!plugin"

    @classmethod
    def from_yaml(cls, loader, node):
        """Build a :class:`~eodag.config.PluginConfig` from Yaml"""
        cls.validate(tuple(node_key.value for node_key, _ in node.value))
        return loader.construct_yaml_object(node, cls)

    @classmethod
    def from_mapping(cls, mapping):
        """Build a :class:`~eodag.config.PluginConfig` from a mapping"""
        c = cls()
        c.__dict__.update(mapping)
        return c

    @staticmethod
    def validate(config_keys):
        """Validate a :class:`~eodag.config.PluginConfig`"""
        if "type" not in config_keys:
            raise ValidationError(
                "A Plugin config must specify the Plugin it configures"
            )

    def update(self, mapping):
        """Update the configuration parameters with values from `mapping`

        :param dict mapping: The mapping from which to override configuration parameters
        """
        if mapping is None:
            mapping = {}
        merge_mappings(
            self.__dict__, {k: v for k, v in mapping.items() if v is not None}
        )


def load_default_config():
    """Build the providers configuration into a dictionnary

    :returns: The default provider's configuration
    :rtype: dict
    """
    config = {}
    with open(resource_filename("eodag", "resources/providers.yml"), "r") as fh:
        try:
            # Providers configs are stored in this file as separated yaml documents
            # Load all of it
            providers_configs = yaml.load_all(fh, Loader=yaml.Loader)
        except yaml.parser.ParserError as e:
            logger.error("Unable to load configuration")
            raise e
        for provider_config in providers_configs:
            # For all providers, set the default outputs_prefix of its download plugin
            # as tempdir in a portable way
            for param_name in ("download", "api"):
                if param_name in vars(provider_config):
                    param_value = getattr(provider_config, param_name)
                    param_value.outputs_prefix = tempfile.gettempdir()
            # Set default priority to 0
            provider_config.__dict__.setdefault("priority", 0)
            config[provider_config.name] = provider_config
        return config


def override_config_from_file(config, file_path):
    """Override a configuration with the values in a file

    :param dict config: An eodag providers configuration dictionary
    :param file_path: The path to the file from where the new values will be read
    :type file_path: str or unicode
    """
    logger.info("Loading user configuration from: %s", os.path.abspath(file_path))
    with open(os.path.abspath(os.path.realpath(file_path)), "r") as fh:
        try:
            config_in_file = yaml.safe_load(fh)
            if config_in_file is None:
                return
            utf8_everywhere(config_in_file)
        except yaml.parser.ParserError as e:
            logger.error("Unable to load user configuration file")
            raise e
    override_config_from_mapping(config, config_in_file)


def override_config_from_env(config):
    """Override a configuration with environment variables values

    :param dict config: An eodag providers configuration dictionary
    """

    def build_mapping_from_env(env_var, env_value, mapping):
        """Recursively build a dictionary from an environment variable.

        The environment variable must respect the pattern: KEY1__KEY2__[...]__KEYN.
        It will be transformed to::

            {
                "key1": {
                    "key2": {
                        {...}
                    }
                }
            }

        :param env_var: The environment variable to be transformed into a dictionary
        :type env_var: str or unicode
        :param env_value: The value from environment variable
        :type env_value: str or unicode
        :param dict mapping: The mapping in which the value will be created
        """
        parts = env_var.split("__")
        if len(parts) == 1:
            mapping[parts[0]] = env_value
        else:
            new_map = mapping.setdefault(parts[0], {})
            build_mapping_from_env("__".join(parts[1:]), env_value, new_map)

    mapping_from_env = {}
    for env_var in os.environ:
        if env_var.startswith("EODAG__"):
            build_mapping_from_env(
                env_var[len("EODAG__") :].lower(),  # noqa
                os.environ[env_var],
                mapping_from_env,
            )

    override_config_from_mapping(config, mapping_from_env)


def override_config_from_mapping(config, mapping):
    """Override a configuration with the values in a mapping

    :param dict config: An eodag providers configuration dictionary
    :param dict mapping: The mapping containing the values to be overriden
    """
    for provider, new_conf in mapping.items():
        old_conf = config.get(provider)
        if old_conf is not None:
            old_conf.update(new_conf)
        else:
            mapping[provider] = ProviderConfig.from_mapping(new_conf)



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function, unicode_literals

import logging
import os
import re

import numpy
import rasterio
import requests
import six
import xarray as xr
from rasterio.enums import Resampling
from rasterio.vrt import WarpedVRT
from requests import HTTPError
from shapely import geometry, geos, wkb, wkt

from eodag.api.product.drivers import DRIVERS, NoDriver
from eodag.api.product.metadata_mapping import NOT_AVAILABLE, NOT_MAPPED
from eodag.utils import ProgressCallback
from eodag.utils.exceptions import DownloadError, UnsupportedDatasetAddressScheme

try:
    from shapely.errors import TopologicalError
except ImportError:
    from shapely.geos import TopologicalError


logger = logging.getLogger("eodag.api.product")


class EOProduct(object):
    """A wrapper around an Earth Observation Product originating from a search.

    Every Search plugin instance must build an instance of this class for each of
    the result of its query method, and return a list of such instances. A EOProduct
    has a `location` attribute that initially points to its remote location, but is
    later changed to point to its path on the filesystem when the product has been
    downloaded. It also has a `remote_location` that always points to the remote
    location, so that the product can be downloaded at anytime if it is deleted from
    the filesystem. An EOProduct instance also has a reference to the search
    parameters that led to its creation.

    :param provider: The provider from which the product originates
    :type provider: str or unicode
    :param properties: The metadata of the product
    :type properties: dict

    .. note::
        The geojson spec `enforces <https://github.com/geojson/draft-geojson/pull/6>`_
        the expression of geometries as
        WGS84 CRS (EPSG:4326) coordinates and EOProduct is intended to be transmitted
        as geojson between applications. Therefore it stores geometries in the before
        mentioned CRS.
    """

    def __init__(self, provider, properties, *args, **kwargs):
        self.provider = provider
        self.product_type = kwargs.get("productType")
        self.location = self.remote_location = properties.get("downloadLink", "")
        self.properties = {
            key: value
            for key, value in properties.items()
            if key != "geometry" and value not in [NOT_MAPPED, NOT_AVAILABLE]
        }
        product_geometry = properties["geometry"]
        # Best effort to understand provider specific geometry (the default is to
        # assume an object implementing the Geo Interface: see
        # https://gist.github.com/2217756)
        if isinstance(product_geometry, six.string_types):
            try:
                product_geometry = wkt.loads(product_geometry)
            except geos.WKTReadingError:
                try:
                    product_geometry = wkb.loads(product_geometry)
                # Also catching TypeError because product_geometry can be a unicode
                # string and not a bytes string
                except (geos.WKBReadingError, TypeError):
                    # Let's try 'latmin lonmin latmax lonmax'
                    coords = [float(coord) for coord in product_geometry.split(" ")]
                    if len(coords) == 4:
                        product_geometry = geometry.box(
                            coords[1], coords[0], coords[3], coords[2]
                        )
                    else:
                        # Giv up!
                        raise
        self.geometry = self.search_intersection = geometry.shape(product_geometry)
        self.search_args = args
        self.search_kwargs = kwargs
        if self.search_kwargs.get("geometry") is not None:
            searched_bbox = self.search_kwargs["geometry"]
            searched_bbox_as_shape = geometry.box(
                searched_bbox["lonmin"],
                searched_bbox["latmin"],
                searched_bbox["lonmax"],
                searched_bbox["latmax"],
            )
            try:
                self.search_intersection = self.geometry.intersection(
                    searched_bbox_as_shape
                )
            except TopologicalError:
                import traceback as tb

                logger.warning(
                    "Unable to intersect the requested extent: %s with the product "
                    "geometry: %s. Got:\n%s",
                    searched_bbox_as_shape,
                    product_geometry,
                    tb.format_exc(),
                )
                self.search_intersection = None
        self.driver = DRIVERS.get(self.product_type, NoDriver())
        self.downloader = None
        self.downloader_auth = None

    def get_data(self, crs, resolution, band, extent):
        """Retrieves all or part of the raster data abstracted by the :class:`EOProduct`

        :param crs: The coordinate reference system in which the dataset should be
                    returned
        :type crs: str or unicode
        :param resolution: The resolution in which the dataset should be returned
                           (given in the unit of the crs)
        :type resolution: float
        :param band: The band of the dataset to retrieve (e.g.: 'B01')
        :type band: str or unicode
        :param extent: The coordinates on which to zoom as a tuple
                       (min_x, min_y, max_x, max_y) in the given `crs`
        :type extent: (float, float, float, float)
        :returns: The numeric matrix corresponding to the sub dataset or an empty
                  array if unable to get the data
        :rtype: xarray.DataArray
        """
        fail_value = xr.DataArray(numpy.empty(0))
        try:
            dataset_address = self.driver.get_data_address(self, band)
        except UnsupportedDatasetAddressScheme:
            logger.warning(
                "Eodag does not support getting data from distant sources by now. "
                "Falling back to first downloading the product and then getting the "
                "data..."
            )
            try:
                path_of_downloaded_file = self.download()
            except (RuntimeError, DownloadError):
                import traceback

                logger.warning(
                    "Error while trying to download the product:\n %s",
                    traceback.format_exc(),
                )
                logger.warning(
                    "There might be no download plugin registered for this EO product. "
                    "Try performing: product.register_downloader(download_plugin, "
                    "auth_plugin) before trying to call product.get_data(...)"
                )
                return fail_value
            if not path_of_downloaded_file:
                return fail_value
            dataset_address = self.driver.get_data_address(self, band)
        min_x, min_y, max_x, max_y = extent
        height = int((max_y - min_y) / resolution)
        width = int((max_x - min_x) / resolution)
        out_shape = (width, height)
        with rasterio.open(dataset_address) as src:
            with WarpedVRT(src, crs=crs, resampling=Resampling.bilinear) as vrt:
                array = vrt.read(
                    1,
                    window=vrt.window(*extent),
                    out_shape=out_shape,
                    resampling=Resampling.bilinear,
                )
                return xr.DataArray(array)

    def as_dict(self):
        """Builds a representation of EOProduct as a dictionary to enable its geojson
        serialization

        :returns: The representation of a :class:`~eodag.api.product.EOProduct` as a
                  Python dict
        :rtype: dict
        """
        search_intersection = None
        if self.search_intersection is not None:
            search_intersection = geometry.mapping(self.search_intersection)
        geojson_repr = {
            "type": "Feature",
            "geometry": geometry.mapping(self.geometry),
            "id": self.properties["id"],
            "properties": {
                "eodag_product_type": self.product_type,
                "eodag_provider": self.provider,
                "eodag_search_intersection": search_intersection,
            },
        }
        geojson_repr["properties"].update(
            {
                key: value
                for key, value in self.properties.items()
                if key not in ("geometry", "id")
            }
        )
        return geojson_repr

    @classmethod
    def from_geojson(cls, feature):
        """Builds an :class:`~eodag.api.product.EOProduct` object from its
        representation as geojson

        :param feature: The representation of a :class:`~eodag.api.product.EOProduct`
                        as a Python dict
        :type feature: dict
        :returns: An instance of :class:`~eodag.api.product.EOProduct`
        :rtype: :class:`~eodag.api.product.EOProduct`
        """
        properties = feature["properties"]
        properties["geometry"] = feature["geometry"]
        properties["id"] = feature["id"]
        provider = feature["properties"]["eodag_provider"]
        product_type = feature["properties"]["eodag_product_type"]
        obj = cls(provider, properties, productType=product_type)
        obj.search_intersection = geometry.shape(
            feature["properties"]["eodag_search_intersection"]
        )
        return obj

    # Implementation of geo-interface protocol (See
    # https://gist.github.com/sgillies/2217756)
    __geo_interface__ = property(as_dict)

    def __repr__(self):
        return "{}(id={}, provider={})".format(
            self.__class__.__name__, self.properties["id"], self.provider
        )

    def encode(self, raster, encoding="protobuf"):
        """Encode the subset to a network-compatible format.

        :param raster: The raster data to encode
        :type raster: xarray.DataArray
        :param encoding: The encoding of the export
        :type encoding: str or unicode
        :return: The data encoded in the specified encoding
        :rtype: bytes
        """
        # If no encoding return an empty byte
        if not encoding:
            logger.warning("Trying to encode a raster without specifying an encoding")
            return b""
        strategy = getattr(self, "_{encoding}".format(**locals()), None)
        if strategy:
            return strategy(raster)
        logger.error("Unknown encoding: %s", encoding)
        return b""

    def _protobuf(self, raster):
        """Google's Protocol buffers encoding strategy.

        :param raster: The raster to encode
        :type raster: xarray.DataArray
        :returns: The raster data represented by this subset in protocol buffers
                  encoding
        :rtype: bytes
        """
        from eodag.api.product.protobuf import eo_product_pb2

        subdataset = eo_product_pb2.EOProductSubdataset()
        subdataset.id = self.properties["id"]
        subdataset.producer = self.provider
        subdataset.product_type = self.product_type
        subdataset.platform = self.properties["platformSerialIdentifier"]
        subdataset.sensor = self.properties["instrument"]
        data = subdataset.data
        data.array.extend(list(raster.values.flatten().astype(int)))
        data.shape.extend(list(raster.values.shape))
        data.dtype = raster.values.dtype.name
        return subdataset.SerializeToString()

    def register_downloader(self, downloader, authenticator):
        """Give to the product the information needed to download itself.

        :param downloader: The download method that it can use
        :type downloader: Concrete subclass of
                          :class:`~eodag.plugins.download.base.Download` or
                          :class:`~eodag.plugins.api.base.Api`
        :param authenticator: The authentication method needed to perform the download
        :type authenticator: Concrete subclass of
                             :class:`~eodag.plugins.authentication.base.Authentication`
        """
        self.downloader = downloader
        self.downloader_auth = authenticator

    def download(self, progress_callback=None):
        """Download the EO product using the provided download plugin and the
        authenticator if necessary.

        The actual download of the product occurs only at the first call of this
        method. A side effect of this method is that it changes the ``location``
        attribute of an EOProduct, from its remote address to the local address.

        :param progress_callback: (optional) A method or a callable object
                                  which takes a current size and a maximum
                                  size as inputs and handle progress bar
                                  creation and update to give the user a
                                  feedback on the download progress
        :type progress_callback: :class:`~eodag.utils.ProgressCallback` or None
        :returns: The absolute path to the downloaded product on the local filesystem
        :rtype: str or unicode
        :raises: :class:`~eodag.utils.exceptions.PluginImplementationError`
        :raises: :class:`RuntimeError`
        """
        if progress_callback is None:
            progress_callback = ProgressCallback()
        if self.downloader is None:
            raise RuntimeError(
                "EO product is unable to download itself due to lacking of a "
                "download plugin"
            )

        auth = (
            self.downloader_auth.authenticate()
            if self.downloader_auth is not None
            else self.downloader_auth
        )
        # resolve remote location if needed with downloader configuration
        self.remote_location = self.remote_location % vars(self.downloader.config)
        fs_location = self.downloader.download(
            self, auth=auth, progress_callback=progress_callback
        )
        if fs_location is None:
            raise DownloadError(
                "Invalid file location returned by download process: '{}'".format(
                    fs_location
                )
            )
        self.location = "file://{}".format(fs_location)
        logger.debug(
            "Product location updated from '%s' to '%s'",
            self.remote_location,
            self.location,
        )
        logger.info(
            "Remote location of the product is still available through its "
            "'remote_location' property: %s",
            self.remote_location,
        )
        return self.location

    def get_quicklook(self, filename=None, base_dir=None, progress_callback=None):
        """Download the quick look image of a given EOProduct from its provider if it
        exists.

        :param filename: (optional) the name to give to the downloaded quicklook.
        :type filename: str (Python 3) or unicode (Python 2)
        :param base_dir: (optional) the absolute path of the directory where to store
                         the quicklooks in the filesystem. If it is not given, it
                         defaults to the `quicklooks` directory under this EO product
                         downloader's ``outputs_prefix`` config param
        :type base_dir: str (Python 3) or unicode (Python 2)
        :param progress_callback: (optional) A method or a callable object
                                  which takes a current size and a maximum
                                  size as inputs and handle progress bar
                                  creation and update to give the user a
                                  feedback on the download progress
        :type progress_callback: :class:`~eodag.utils.ProgressCallback` or None
        :returns: The absolute path of the downloaded quicklook
        :rtype: str (Python 3) or unicode (Python 2)

        .. versionchanged::
            1.0

                * Added the ``base_dir`` optional parameter to choose where to download
                  the retrieved quicklook
        """

        def format_quicklook_address():
            """If the quicklook address is a Python format string, resolve the
            formatting with the properties of the product."""
            fstrmatch = re.match(r".*{.+}*.*", self.properties["quicklook"])
            if fstrmatch:
                self.properties["quicklook"].format(
                    {
                        prop_key: prop_val
                        for prop_key, prop_val in self.properties.items()
                        if prop_key != "quicklook"
                    }
                )

        if progress_callback is None:
            progress_callback = ProgressCallback()

        if self.properties["quicklook"] is None:
            logger.warning(
                "Missing information to retrieve quicklook for EO product: %s",
                self.properties["id"],
            )
            return ""

        format_quicklook_address()

        if base_dir is not None:
            quicklooks_base_dir = os.path.abspath(os.path.realpath(base_dir))
        else:
            quicklooks_base_dir = os.path.join(
                self.downloader.config.outputs_prefix, "quicklooks"
            )
        if not os.path.isdir(quicklooks_base_dir):
            os.makedirs(quicklooks_base_dir)
        quicklook_file = os.path.join(
            quicklooks_base_dir,
            filename if filename is not None else self.properties["id"],
        )

        if not os.path.isfile(quicklook_file):
            # VERY SPECIAL CASE (introduced by the onda provider): first check if
            # it is a HTTP URL. If not, we assume it is a byte string, in which case
            # we just write the content into the quicklook_file and return it.
            if not (
                self.properties["quicklook"].startswith("http")
                or self.properties["quicklook"].startswith("https")
            ):
                with open(quicklook_file, "wb") as fd:
                    fd.write(self.properties["quicklook"])
                return quicklook_file

            auth = (
                self.downloader_auth.authenticate()
                if self.downloader_auth is not None
                else None
            )
            with requests.get(
                self.properties["quicklook"], stream=True, auth=auth
            ) as stream:
                try:
                    stream.raise_for_status()
                except HTTPError as e:
                    import traceback as tb

                    logger.error("Error while getting resource :\n%s", tb.format_exc())
                    return str(e)
                else:
                    stream_size = int(stream.headers.get("content-length", 0))
                    with open(quicklook_file, "wb") as fhandle:
                        for chunk in stream.iter_content(chunk_size=64 * 1024):
                            if chunk:
                                fhandle.write(chunk)
                                progress_callback(len(chunk), stream_size)
                    logger.info("Download recorded in %s", quicklook_file)
        return quicklook_file



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

import datetime
import logging
import time

import dateutil.parser
from shapely import geometry

from eodag.plugins.crunch.base import Crunch


logger = logging.getLogger('eodag.plugins.crunch.filter_latest')


class FilterLatestIntersect(Crunch):

    @staticmethod
    def sort_product_by_start_date(product):
        start_date = product.properties.get('startTimeFromAscendingNode')
        if not start_date:
            # Retrieve year, month, day, hour, minute, second of EPOCH start
            epoch = time.gmtime(0)[:-3]
            start_date = datetime.datetime(*epoch).isoformat()
        return dateutil.parser.parse(start_date)

    def proceed(self, products, **search_params):
        """Filter latest products (the ones with a the highest start date) that intersect search extent."""
        logger.debug('Start filtering for latest products')
        if not products:
            return []
        # Warning: May crash if startTimeFromAscendingNode is not in the appropriate format
        products.sort(key=self.sort_product_by_start_date, reverse=True)
        filtered = []
        add_to_filtered = filtered.append
        footprint = search_params.get('footprint')
        if not footprint:
            return products
        search_extent = geometry.box(
            footprint['lonmin'],
            footprint['latmin'],
            footprint['lonmax'],
            footprint['latmax']
        )
        logger.debug('Initial requested extent area: %s', search_extent.area)
        for product in products:
            logger.debug('Uncovered extent area: %s', search_extent.area)
            if product.search_intersection:
                logger.debug('Product %r intersects the requested extent. Adding it to the final result', product)
                add_to_filtered(product)
            search_extent = search_extent.difference(product.geometry)
            if search_extent.is_empty:
                logger.debug('The requested extent is now entirely covered by the search result')
                break
        logger.debug('Finished filtering products. Resulting products: %r', filtered)
        return filtered



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from eodag.rpc.protocol import eodag_pb2 as eodag__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


class EOProductTypeStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.ListProductTypes = channel.unary_stream(
        '/eodag.EOProductType/ListProductTypes',
        request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        response_deserializer=eodag__pb2.EOProductTypeSchema.FromString,
        )
    self.SearchProductType = channel.unary_stream(
        '/eodag.EOProductType/SearchProductType',
        request_serializer=eodag__pb2.EOProductTypeSearchRequest.SerializeToString,
        response_deserializer=eodag__pb2.EOProductTypeSchema.FromString,
        )


class EOProductTypeServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def ListProductTypes(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SearchProductType(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_EOProductTypeServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'ListProductTypes': grpc.unary_stream_rpc_method_handler(
          servicer.ListProductTypes,
          request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
          response_serializer=eodag__pb2.EOProductTypeSchema.SerializeToString,
      ),
      'SearchProductType': grpc.unary_stream_rpc_method_handler(
          servicer.SearchProductType,
          request_deserializer=eodag__pb2.EOProductTypeSearchRequest.FromString,
          response_serializer=eodag__pb2.EOProductTypeSchema.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'eodag.EOProductType', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class EOProductStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.SearchProduct = channel.unary_stream(
        '/eodag.EOProduct/SearchProduct',
        request_serializer=eodag__pb2.EOProductSearchRequest.SerializeToString,
        response_deserializer=eodag__pb2.EOProductInfo.FromString,
        )
    self.DownloadProduct = channel.unary_stream(
        '/eodag.EOProduct/DownloadProduct',
        request_serializer=eodag__pb2.EOProductDownloadRequest.SerializeToString,
        response_deserializer=eodag__pb2.NetCDFBuffer.FromString,
        )


class EOProductServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def SearchProduct(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DownloadProduct(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_EOProductServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'SearchProduct': grpc.unary_stream_rpc_method_handler(
          servicer.SearchProduct,
          request_deserializer=eodag__pb2.EOProductSearchRequest.FromString,
          response_serializer=eodag__pb2.EOProductInfo.SerializeToString,
      ),
      'DownloadProduct': grpc.unary_stream_rpc_method_handler(
          servicer.DownloadProduct,
          request_deserializer=eodag__pb2.EOProductDownloadRequest.FromString,
          response_serializer=eodag__pb2.NetCDFBuffer.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'eodag.EOProduct', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: eodag.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='eodag.proto',
  package='eodag',
  syntax='proto3',
  serialized_pb=_b('\n\x0b\x65odag.proto\x12\x05\x65odag\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x19google/protobuf/any.proto\">\n\x1b\x45OProductTypeMetadataSchema\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x12\n\nvalue_type\x18\x02 \x01(\t\"h\n\x13\x45OProductTypeSchema\x12\n\n\x02id\x18\x01 \x01(\t\x12\x30\n\x04meta\x18\x02 \x03(\x0b\x32\".eodag.EOProductTypeMetadataSchema\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\"+\n\x1a\x45OProductTypeSearchRequest\x12\r\n\x05query\x18\x01 \x01(\t\"\xae\x01\n\x16\x45OProductSearchRequest\x12\x14\n\x0cproduct_type\x18\x01 \x01(\t\x12\x11\n\tfootprint\x18\x02 \x01(\t\x12)\n\x05start\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\'\n\x03\x65nd\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x17\n\x0fmetadata_filter\x18\x05 \x01(\t\"<\n\x08Metadata\x12\x0b\n\x03key\x18\x01 \x01(\t\x12#\n\x05value\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any\">\n\rEOProductInfo\x12\n\n\x02id\x18\x01 \x01(\t\x12!\n\x08metadata\x18\x02 \x01(\x0b\x32\x0f.eodag.Metadata\".\n\x18\x45OProductDownloadRequest\x12\x12\n\nproduct_id\x18\x02 \x01(\t\"\x1f\n\x0cNetCDFBuffer\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c\x32\xb3\x01\n\rEOProductType\x12J\n\x10ListProductTypes\x12\x16.google.protobuf.Empty\x1a\x1a.eodag.EOProductTypeSchema\"\x00\x30\x01\x12V\n\x11SearchProductType\x12!.eodag.EOProductTypeSearchRequest\x1a\x1a.eodag.EOProductTypeSchema\"\x00\x30\x01\x32\xa2\x01\n\tEOProduct\x12H\n\rSearchProduct\x12\x1d.eodag.EOProductSearchRequest\x1a\x14.eodag.EOProductInfo\"\x00\x30\x01\x12K\n\x0f\x44ownloadProduct\x12\x1f.eodag.EOProductDownloadRequest\x1a\x13.eodag.NetCDFBuffer\"\x00\x30\x01\x62\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,google_dot_protobuf_dot_any__pb2.DESCRIPTOR,])




_EOPRODUCTTYPEMETADATASCHEMA = _descriptor.Descriptor(
  name='EOProductTypeMetadataSchema',
  full_name='eodag.EOProductTypeMetadataSchema',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='eodag.EOProductTypeMetadataSchema.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value_type', full_name='eodag.EOProductTypeMetadataSchema.value_type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=111,
  serialized_end=173,
)


_EOPRODUCTTYPESCHEMA = _descriptor.Descriptor(
  name='EOProductTypeSchema',
  full_name='eodag.EOProductTypeSchema',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='eodag.EOProductTypeSchema.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='meta', full_name='eodag.EOProductTypeSchema.meta', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='eodag.EOProductTypeSchema.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=175,
  serialized_end=279,
)


_EOPRODUCTTYPESEARCHREQUEST = _descriptor.Descriptor(
  name='EOProductTypeSearchRequest',
  full_name='eodag.EOProductTypeSearchRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='query', full_name='eodag.EOProductTypeSearchRequest.query', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=281,
  serialized_end=324,
)


_EOPRODUCTSEARCHREQUEST = _descriptor.Descriptor(
  name='EOProductSearchRequest',
  full_name='eodag.EOProductSearchRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='product_type', full_name='eodag.EOProductSearchRequest.product_type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='footprint', full_name='eodag.EOProductSearchRequest.footprint', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='start', full_name='eodag.EOProductSearchRequest.start', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='end', full_name='eodag.EOProductSearchRequest.end', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='metadata_filter', full_name='eodag.EOProductSearchRequest.metadata_filter', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=327,
  serialized_end=501,
)


_METADATA = _descriptor.Descriptor(
  name='Metadata',
  full_name='eodag.Metadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='eodag.Metadata.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='eodag.Metadata.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=503,
  serialized_end=563,
)


_EOPRODUCTINFO = _descriptor.Descriptor(
  name='EOProductInfo',
  full_name='eodag.EOProductInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='eodag.EOProductInfo.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='eodag.EOProductInfo.metadata', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=565,
  serialized_end=627,
)


_EOPRODUCTDOWNLOADREQUEST = _descriptor.Descriptor(
  name='EOProductDownloadRequest',
  full_name='eodag.EOProductDownloadRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='product_id', full_name='eodag.EOProductDownloadRequest.product_id', index=0,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=629,
  serialized_end=675,
)


_NETCDFBUFFER = _descriptor.Descriptor(
  name='NetCDFBuffer',
  full_name='eodag.NetCDFBuffer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='content', full_name='eodag.NetCDFBuffer.content', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=677,
  serialized_end=708,
)

_EOPRODUCTTYPESCHEMA.fields_by_name['meta'].message_type = _EOPRODUCTTYPEMETADATASCHEMA
_EOPRODUCTSEARCHREQUEST.fields_by_name['start'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_EOPRODUCTSEARCHREQUEST.fields_by_name['end'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_METADATA.fields_by_name['value'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_EOPRODUCTINFO.fields_by_name['metadata'].message_type = _METADATA
DESCRIPTOR.message_types_by_name['EOProductTypeMetadataSchema'] = _EOPRODUCTTYPEMETADATASCHEMA
DESCRIPTOR.message_types_by_name['EOProductTypeSchema'] = _EOPRODUCTTYPESCHEMA
DESCRIPTOR.message_types_by_name['EOProductTypeSearchRequest'] = _EOPRODUCTTYPESEARCHREQUEST
DESCRIPTOR.message_types_by_name['EOProductSearchRequest'] = _EOPRODUCTSEARCHREQUEST
DESCRIPTOR.message_types_by_name['Metadata'] = _METADATA
DESCRIPTOR.message_types_by_name['EOProductInfo'] = _EOPRODUCTINFO
DESCRIPTOR.message_types_by_name['EOProductDownloadRequest'] = _EOPRODUCTDOWNLOADREQUEST
DESCRIPTOR.message_types_by_name['NetCDFBuffer'] = _NETCDFBUFFER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

EOProductTypeMetadataSchema = _reflection.GeneratedProtocolMessageType('EOProductTypeMetadataSchema', (_message.Message,), dict(
  DESCRIPTOR = _EOPRODUCTTYPEMETADATASCHEMA,
  __module__ = 'eodag_pb2'
  # @@protoc_insertion_point(class_scope:eodag.EOProductTypeMetadataSchema)
  ))
_sym_db.RegisterMessage(EOProductTypeMetadataSchema)

EOProductTypeSchema = _reflection.GeneratedProtocolMessageType('EOProductTypeSchema', (_message.Message,), dict(
  DESCRIPTOR = _EOPRODUCTTYPESCHEMA,
  __module__ = 'eodag_pb2'
  # @@protoc_insertion_point(class_scope:eodag.EOProductTypeSchema)
  ))
_sym_db.RegisterMessage(EOProductTypeSchema)

EOProductTypeSearchRequest = _reflection.GeneratedProtocolMessageType('EOProductTypeSearchRequest', (_message.Message,), dict(
  DESCRIPTOR = _EOPRODUCTTYPESEARCHREQUEST,
  __module__ = 'eodag_pb2'
  # @@protoc_insertion_point(class_scope:eodag.EOProductTypeSearchRequest)
  ))
_sym_db.RegisterMessage(EOProductTypeSearchRequest)

EOProductSearchRequest = _reflection.GeneratedProtocolMessageType('EOProductSearchRequest', (_message.Message,), dict(
  DESCRIPTOR = _EOPRODUCTSEARCHREQUEST,
  __module__ = 'eodag_pb2'
  # @@protoc_insertion_point(class_scope:eodag.EOProductSearchRequest)
  ))
_sym_db.RegisterMessage(EOProductSearchRequest)

Metadata = _reflection.GeneratedProtocolMessageType('Metadata', (_message.Message,), dict(
  DESCRIPTOR = _METADATA,
  __module__ = 'eodag_pb2'
  # @@protoc_insertion_point(class_scope:eodag.Metadata)
  ))
_sym_db.RegisterMessage(Metadata)

EOProductInfo = _reflection.GeneratedProtocolMessageType('EOProductInfo', (_message.Message,), dict(
  DESCRIPTOR = _EOPRODUCTINFO,
  __module__ = 'eodag_pb2'
  # @@protoc_insertion_point(class_scope:eodag.EOProductInfo)
  ))
_sym_db.RegisterMessage(EOProductInfo)

EOProductDownloadRequest = _reflection.GeneratedProtocolMessageType('EOProductDownloadRequest', (_message.Message,), dict(
  DESCRIPTOR = _EOPRODUCTDOWNLOADREQUEST,
  __module__ = 'eodag_pb2'
  # @@protoc_insertion_point(class_scope:eodag.EOProductDownloadRequest)
  ))
_sym_db.RegisterMessage(EOProductDownloadRequest)

NetCDFBuffer = _reflection.GeneratedProtocolMessageType('NetCDFBuffer', (_message.Message,), dict(
  DESCRIPTOR = _NETCDFBUFFER,
  __module__ = 'eodag_pb2'
  # @@protoc_insertion_point(class_scope:eodag.NetCDFBuffer)
  ))
_sym_db.RegisterMessage(NetCDFBuffer)



_EOPRODUCTTYPE = _descriptor.ServiceDescriptor(
  name='EOProductType',
  full_name='eodag.EOProductType',
  file=DESCRIPTOR,
  index=0,
  options=None,
  serialized_start=711,
  serialized_end=890,
  methods=[
  _descriptor.MethodDescriptor(
    name='ListProductTypes',
    full_name='eodag.EOProductType.ListProductTypes',
    index=0,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_EOPRODUCTTYPESCHEMA,
    options=None,
  ),
  _descriptor.MethodDescriptor(
    name='SearchProductType',
    full_name='eodag.EOProductType.SearchProductType',
    index=1,
    containing_service=None,
    input_type=_EOPRODUCTTYPESEARCHREQUEST,
    output_type=_EOPRODUCTTYPESCHEMA,
    options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_EOPRODUCTTYPE)

DESCRIPTOR.services_by_name['EOProductType'] = _EOPRODUCTTYPE


_EOPRODUCT = _descriptor.ServiceDescriptor(
  name='EOProduct',
  full_name='eodag.EOProduct',
  file=DESCRIPTOR,
  index=1,
  options=None,
  serialized_start=893,
  serialized_end=1055,
  methods=[
  _descriptor.MethodDescriptor(
    name='SearchProduct',
    full_name='eodag.EOProduct.SearchProduct',
    index=0,
    containing_service=None,
    input_type=_EOPRODUCTSEARCHREQUEST,
    output_type=_EOPRODUCTINFO,
    options=None,
  ),
  _descriptor.MethodDescriptor(
    name='DownloadProduct',
    full_name='eodag.EOProduct.DownloadProduct',
    index=1,
    containing_service=None,
    input_type=_EOPRODUCTDOWNLOADREQUEST,
    output_type=_NETCDFBUFFER,
    options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_EOPRODUCT)

DESCRIPTOR.services_by_name['EOProduct'] = _EOPRODUCT

# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, print_function, unicode_literals

import logging
from operator import attrgetter

import pkg_resources

from eodag.plugins.apis.base import Api
from eodag.plugins.authentication.base import Authentication
from eodag.plugins.base import EODAGPluginMount
from eodag.plugins.crunch.base import Crunch
from eodag.plugins.download.base import Download
from eodag.plugins.search.base import Search
from eodag.utils.exceptions import UnsupportedProductType, UnsupportedProvider

logger = logging.getLogger("eodag.plugins.manager")


class PluginManager(object):
    """A manager for the plugins.

    The role of instances of this class (normally only one instance exists,
    created during instantiation of :class:`~eodag.api.core.EODataAccessGateway`.
    But nothing is done to enforce this) is to instantiate the plugins
    according to the providers configuration, keep track of them in memory, and
    manage a cache of plugins already constructed. The providers configuration contains
    information such as the name of the provider, the internet endpoint for accessing
    it, and the plugins to use to perform defined actions (search, download,
    authenticate, crunch).

    :param providers_config: The configuration with all information about the providers
                             supported by the `eodag`
    :type providers_config: dict
    """

    supported_topics = {"search", "download", "crunch", "auth", "api"}

    def __init__(self, providers_config):
        self.providers_config = providers_config
        # Load all the plugins. This will make all plugin classes of a particular
        # type to be available in the base plugin class's 'plugins' attribute.
        # For example, by importing module 'eodag.plugins.search.resto', the plugin
        # 'RestoSearch' will be available in self.supported_topics['search'].plugins
        for topic in self.supported_topics:
            # This way of discovering plugins means that anyone can create eodag
            # plugins as a separate python package (though it must require eodag), and
            # have it discovered as long as they declare an entry point of the type
            # 'eodag.plugins.search' for example in its setup script. See the setup
            # script of eodag for an example of how to do this.
            for entry_point in pkg_resources.iter_entry_points(
                "eodag.plugins.{}".format(topic)
            ):
                try:
                    entry_point.load()
                except ImportError:
                    import traceback as tb

                    logger.warning("Unable to load plugin: %s.", entry_point.name)
                    logger.warning("Reason:\n%s", tb.format_exc())
                    logger.warning(
                        "Check that the plugin module (%s) is importable",
                        entry_point.module_name,
                    )
        self.product_type_to_provider_config_map = {}
        for provider_config in providers_config.values():
            for product_type in provider_config.products:
                product_type_providers = self.product_type_to_provider_config_map.setdefault(  # noqa
                    product_type, []
                )
                product_type_providers.append(provider_config)
                product_type_providers.sort(key=attrgetter("priority"), reverse=True)
        self._built_plugins_cache = {}

    def get_search_plugins(self, product_type=None, provider=None):
        """Build and return all the search plugins supporting the given product type,
        ordered by highest priority, or the search plugin of the given provider

        :param product_type: (Optional) The product type that the constructed plugins
                             must support
        :type product_type: str or unicode
        :param provider: (Optional) The provider on which to get the search plugin
        :type provider: str or unicode
        :returns: All the plugins supporting the product type, one by one (a generator
                  object)
        :rtype: types.GeneratorType(:class:`~eodag.plugins.search.Search`)
        :raises: :class:`~eodag.utils.exceptions.UnsupportedProvider`
        :raises: :class:`~eodag.utils.exceptions.UnsupportedProductType`
        :raises: StopIteration

        .. versionchanged::
            1.0

                * ``product_type`` is now optional. If no product type is provided,
                  return all search plugins, ordered by priority
                * A new optional parameter ``provider`` which defaults to ``None``, if
                  we want to build the search plugin of that provider
        """

        def get_plugin():
            try:
                config.search.products = config.products
                config.search.priority = config.priority
                plugin = self._build_plugin(config.name, config.search, Search)
            except AttributeError:
                config.api.products = config.products
                config.api.priority = config.priority
                plugin = self._build_plugin(config.name, config.api, Api)
            return plugin

        if provider is not None:
            try:
                config = self.providers_config[provider]
            except KeyError:
                raise UnsupportedProvider
            yield get_plugin()
            # Signal the end of iteration as we already have what we wanted
            # In a for-loop, this exception is automatically catched
            raise StopIteration

        if product_type is None:
            for config in sorted(
                self.providers_config.values(), key=attrgetter("priority"), reverse=True
            ):
                yield get_plugin()
            # Signal the end of iteration as we already have what we wanted
            # In a for-loop, this exception is automatically catched
            raise StopIteration
        try:
            for config in self.product_type_to_provider_config_map[product_type]:
                yield get_plugin()
        except KeyError:
            raise UnsupportedProductType(product_type)

    def get_download_plugin(self, product):
        """Build and return the download plugin capable of downloading the given
        product.

        :param product: The product to get a download plugin for
        :type product: :class:`~eodag.api.product._product.EOProduct`
        :returns: The download plugin capable of downloading the product
        :rtype: :class:`~eodag.plugins.download.Download`
        """
        plugin_conf = self.providers_config[product.provider]
        try:
            plugin_conf.download.priority = plugin_conf.priority
            plugin = self._build_plugin(
                product.provider, plugin_conf.download, Download
            )
            return plugin
        except AttributeError:
            plugin_conf.api.priority = plugin_conf.priority
            plugin = self._build_plugin(product.provider, plugin_conf.api, Api)
            return plugin

    def get_auth_plugin(self, provider):
        """Build and return the authentication plugin for the given product_type and
        provider

        :param provider: The provider for which to get the authentication plugin
        :type provider: str or unicode
        :returns: The Authentication plugin for the provider
        :rtype: :class:`~eodag.plugins.authentication.Authentication`

        .. versionchanged::
            1.0

                * ``product_type`` is no longer needed to find the auth plugin
        """
        plugin_conf = self.providers_config[provider]
        try:
            plugin_conf.auth.priority = plugin_conf.priority
            plugin = self._build_plugin(provider, plugin_conf.auth, Authentication)
            return plugin
        except AttributeError:
            # We guess the plugin being built is of type Api, therefore no need
            # for an Auth plugin.
            return None

    @staticmethod
    def get_crunch_plugin(name, **options):
        """Instantiate a eodag Crunch plugin whom class name is `name`, and configure
        it with the `options`

        :param name: The name of the Crunch plugin to instantiate
        :type name: str or unicode
        :param options: The configuration parameters of the cruncher
        :type options: dict
        :return: The cruncher named `name`
        :rtype: :class:`~eodag.plugins.crunch.Crunch`
        """
        Klass = Crunch.get_plugin_by_class_name(name)
        return Klass(options)

    def set_priority(self, provider, priority):
        """Set the priority of the given provider

        :param provider: The provider which is assigned the priority
        :type provider: str or unicode
        :param priority: The priority to assign to the provider
        :type priority: int
        """
        # Update the priority in the configurations so that it is taken into account
        # when a plugin of this provider is latterly built
        for (
            product_type,
            provider_configs,
        ) in self.product_type_to_provider_config_map.items():
            for config in provider_configs:
                if config.name == provider:
                    config.priority = priority
            # Sort the provider configs, taking into account the new priority order
            provider_configs.sort(key=attrgetter("priority"), reverse=True)
        # Update the priority of already built plugins of the given provider
        for provider_name, topic_class in self._built_plugins_cache:
            if provider_name == provider:
                self._built_plugins_cache[(provider, topic_class)].priority = priority

    def _build_plugin(self, provider, plugin_conf, topic_class):
        """Build the plugin of the given topic with the given plugin configuration and
        registered as the given provider

        :param provider: The provider for which to build the plugin
        :type provider: str or unicode
        :param plugin_conf: The configuration of the plugin to be built
        :type plugin_conf: :class:`~eodag.config.PluginConfig`
        :param topic_class: The type of plugin to build
        :type topic_class: :class:`~eodag.plugin.base.PluginTopic`
        :returns: The built plugin
        :rtype: :class:`~eodag.plugin.search.Search` or
                :class:`~eodag.plugin.download.Download` or
                :class:`~eodag.plugin.authentication.Authentication` or
                :class:`~eodag.plugin.crunch.Crunch`
        """
        cached_instance = self._built_plugins_cache.setdefault(
            (provider, topic_class.__name__), None
        )
        if cached_instance is not None:
            return cached_instance
        plugin_class = EODAGPluginMount.get_plugin_by_class_name(
            topic_class, getattr(plugin_conf, "type")
        )
        plugin = plugin_class(provider, plugin_conf)
        self._built_plugins_cache[(provider, topic_class.__name__)] = plugin
        return plugin



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

import logging

from shapely import geometry

try:
    from shapely.errors import TopologicalError
except ImportError:
    from shapely.geos import TopologicalError

from eodag.plugins.crunch.base import Crunch


logger = logging.getLogger('eodag.plugins.crunch.filter_overlap')


class FilterOverlap(Crunch):

    def proceed(self, products, **search_params):
        """Filter products, retaining only those that are overlapping with the search_extent"""
        logger.info('Start filtering for overlapping products')
        filtered = []
        add_to_filtered = filtered.append
        footprint = search_params.get('geometry')
        if not footprint:
            return products
        minimum_overlap = float(self.config.get('minimum_overlap', '0'))
        logger.info('Minimum overlap is: {} %'.format(minimum_overlap))
        search_extent = geometry.box(footprint['lonmin'], footprint['latmin'],
                                     footprint['lonmax'], footprint['latmax'])
        logger.info('Initial requested extent area: %s', search_extent.area)
        if search_extent.area == 0:
            logger.info('No product can overlap a requested extent that is not a polygon (i.e with area=0)')
        else:
            for product in products:
                logger.info('Uncovered extent area: %s', search_extent.area)
                if product.search_intersection:
                    intersection = product.search_intersection
                    product_geometry = product.geometry
                else:  # Product geometry may be invalid
                    if not product.geometry.is_valid:
                        logger.debug('Trying our best to deal with invalid geometry on product: %r', product)
                        product_geometry = product.geometry.buffer(0)
                        try:
                            intersection = search_extent.intersection(product_geometry)
                        except TopologicalError:
                            logger.debug('Product geometry still invalid. Overlap test restricted to containment')
                            if search_extent.contains(product_geometry):
                                logger.debug('Product %r overlaps the search extent. Adding it to filtered results')
                                add_to_filtered(product)
                            continue
                    else:
                        product_geometry = product.geometry
                        intersection = search_extent.intersection(product_geometry)
                ipos = (intersection.area / search_extent.area) * 100
                ipop = (intersection.area / product_geometry.area) * 100
                logger.info('Intersection of product extent and search extent covers %f percent of the search extent '
                            'area', ipos)
                logger.info('Intersection of product extent and search extent covers %f percent of the product extent '
                            'area', ipop)
                if any((search_extent.contains(product.geometry), ipos >= minimum_overlap, ipop >= minimum_overlap)):
                    logger.info('Product %r overlaps the search extent by the specified constraint. Adding it to '
                                'filtered results', product)
                    add_to_filtered(product)
                else:
                    logger.info('Product %r does not overlaps the search extent by the specified constraint. '
                                'Skipping it', product)
        logger.info('Finished filtering products. Resulting products: %r', filtered)
        return filtered



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

import re
import string
from random import SystemRandom

import requests
from lxml import etree
from requests.auth import AuthBase

from eodag.plugins.authentication import Authentication
from eodag.utils import parse_qs, repeatfunc, urlencode, urlparse, urlunparse
from eodag.utils.exceptions import AuthenticationError, MisconfiguredError


class OIDCAuthorizationCodeFlowAuth(Authentication):
    """Implement the authorization code flow of the OpenIDConnect authorization specification.

    The `OpenID Connect <http://openid.net/specs/openid-connect-core-1_0.html>`_ specification
    adds an authentication layer on top of oauth 2.0. This plugin implements the
    `authorization code flow <http://openid.net/specs/openid-connect-core-1_0.html#Authentication>`_
    option of this specification.
    The particularity of this plugin is that it proceeds to a headless (not involving the user)
    interaction with the OpenID provider (if necessary) to authenticate a
    registered user with its username and password on the server and then granting to eodag the
    necessary rights. It does that using the client ID of the eodag provider that use it.
    If the client secret of the eodag provider using this plugin is known, it is used in conjunction
    with the client ID to do a BASIC Auth during the token exchange request.
    The headless interaction is fully configurable, and rely on XPATH to retrieve all the necessary
    information.

    The configuration keys of this plugin are as follows (they have no defaults)::

        # (mandatory) The authorization url of the server (where to query for grants)
        authorization_uri:

        # (mandatory) The callback url that will handle the code given by the OIDC provider
        redirect_uri:

        # (mandatory) The url to query to exchange the authorization code obtained from the OIDC provider
        # for an authorized token
        token_uri:

        # (mandatory) The OIDC provider's client ID of the eodag provider
        client_id:

        # (mandatory) Wether a user consent is needed during the authentication
        user_consent_needed:

        # (mandatory) One of: json, data or params. This is the way to pass the data to the POST request
        # that is made to the token server. They correspond to the recognised keywords arguments
        # of the Python `requests <http://docs.python-requests.org/>`_ library
        token_exchange_post_data_method:

        # (mandatory) The key pointing to the token in the json response to the POST request to the token server
        token_key:

        # (mandatory) One of qs or header. This is how the token obtained will be used to authenticate the user
        # on protected requests. If 'qs' is chosen, then 'token_qs_key' is mandatory
        token_provision:

        # (mandatory) The xpath to the HTML form element representing the user login form
        login_form_xpath:

        # (mandatory) Where to look for the authentication_uri. One of 'config' (in the configuration) or 'login-form'
        # (use the 'action' URL found in the login form retrieved with login_form_xpath). If the value is 'config',
        # authentication_uri config param is mandatory
        authentication_uri_source:

        # (optional) The URL of the authentication backend of the OIDC provider
        authentication_uri:

        # (optional) The xpath to the user consent form. The form is searched in the content of the response
        # to the authorization request
        user_consent_form_xpath:

        # (optional) The data that will be passed with the POST request on the form 'action' URL. The data are
        # given as a key value pairs, the keys representing the data key and the value being either
        # a 'constant' string value, or a string of the form 'xpath(<path-to-a-value-to-be-retrieved>)'
        # and representing a value to be retrieved in the user consent form. The xpath must resolve
        # directly to a string value, not to an HTML element. Example:
        # 'xpath(//input[@name="sessionDataKeyConsent"]/@value)'
        user_consent_form_data:

        # (optional) A mapping giving additional data to be passed to the login POST request. The value follows the
        # same rules as with user_consent_form_data
        additional_login_form_data:

        # (optional) The OIDC provider's client secret of the eodag provider
        client_secret:

        # (optional) A mapping between OIDC url query string and token handler query string
        # params (only necessary if they are not the same as for OIDC). This is eodag provider
        # dependant
        token_exchange_params:
          redirect_uri:
          client_id:

        # (optional) Only necessary when 'token_provision' is 'qs'. Refers to the name of the query param to be
        # used in the query request
        token_qs_key:

    """
    SCOPE = 'openid'
    RESPONSE_TYPE = 'code'
    CONFIG_XPATH_REGEX = re.compile(r'^xpath\((?P<xpath_value>.+)\)$')

    def __init__(self, provider, config):
        super(OIDCAuthorizationCodeFlowAuth, self).__init__(provider, config)
        if getattr(self.config, 'token_provision', None) not in ('qs', 'header'):
            raise MisconfiguredError('Provider config parameter "token_provision" must be one of "qs" or "header"')
        if self.config.token_provision == 'qs' and not getattr(self.config, 'token_qs_key', ''):
            raise MisconfiguredError('Provider config parameter "token_provision" with value "qs" must have '
                                     '"token_qs_key" config parameter as well')
        self.session = requests.Session()

    def authenticate(self):
        state = self.compute_state()
        authentication_response = self.authenticate_user(state)
        exchange_url = authentication_response.url
        if self.config.user_consent_needed:
            user_consent_response = self.grant_user_consent(authentication_response)
            exchange_url = user_consent_response.url
        try:
            token = self.exchange_code_for_token(exchange_url, state)
        except Exception:
            import traceback as tb
            raise AuthenticationError(
                'Something went wrong while trying to get authorization token:\n{}'.format(tb.format_exc())
            )
        return CodeAuthorizedAuth(token, self.config.token_provision, key=getattr(self.config, 'token_qs_key', None))

    def authenticate_user(self, state):
        params = {
            'client_id': self.config.client_id,
            'response_type': self.RESPONSE_TYPE,
            'scope': self.SCOPE,
            'state': state,
            'redirect_uri': self.config.redirect_uri,
        }
        authorization_response = self.session.get(self.config.authorization_uri, params=params)

        login_document = etree.HTML(authorization_response.text)
        login_form = login_document.xpath(self.config.login_form_xpath)[0]
        try:
            # Get the form data to pass to the login form from config or from the login form
            login_data = {
                key: self._constant_or_xpath_extracted(value, login_form)
                for key, value in getattr(self.config, 'additional_login_form_data', {}).items()
            }
            # Add the credentials
            login_data.update(self.config.credentials)
            auth_uri = getattr(self.config, 'authentication_uri', None)
            # Retrieve the authentication_uri from the login form if so configured
            if self.config.authentication_uri_source == 'login-form':
                # Given that the login_form_xpath resolves to an HTML element, if suffices to add '/@action' to get
                # the value of its action attribute to this xpath
                auth_uri = login_form.xpath(self.config.login_form_xpath.rstrip('/') + '/@action')[0]
            return self.session.post(auth_uri, data=login_data)
        except AttributeError as err:
            if 'credentials' in err.args:
                raise MisconfiguredError('Missing Credentials for provider: %s', self.provider)

    def grant_user_consent(self, authentication_response):
        user_consent_document = etree.HTML(authentication_response.text)
        user_consent_form = user_consent_document.xpath(self.config.user_consent_form_xpath)[0]
        # Get the form data to pass to the consent form from config or from the consent form
        user_consent_data = {
            key: self._constant_or_xpath_extracted(value, user_consent_form)
            for key, value in self.config.user_consent_form_data.items()
        }
        return self.session.post(self.config.authorization_uri, data=user_consent_data)

    def exchange_code_for_token(self, authorized_url, state):
        qs = parse_qs(urlparse(authorized_url).query)
        if qs['state'][0] != state:
            raise AuthenticationError(
                'The state received in the authorized url does not match initially computed state')
        code = qs['code'][0]
        token_exchange_data = {
            'redirect_uri': self.config.redirect_uri,
            'client_id': self.config.client_id,
            'code': code,
            'state': state,
        }
        # If necessary, change the keys of the form data that will be passed to the token exchange POST request
        custom_token_exchange_params = getattr(self.config, 'token_exchange_params', {})
        if custom_token_exchange_params:
            token_exchange_data[custom_token_exchange_params['redirect_uri']] = token_exchange_data.pop('redirect_uri')
            token_exchange_data[custom_token_exchange_params['client_id']] = token_exchange_data.pop('client_id')
        # If the client_secret is known, the token exchange request must be authenticated with a BASIC Auth, using the
        # client_id and client_secret as username and password respectively
        if getattr(self.config, 'client_secret', None):
            token_exchange_data.update({
                'auth': (self.config.client_id, self.config.client_secret),
                'grant_type': 'authorization_code',
                'client_secret': self.config.client_secret
            })
        post_request_kwargs = {self.config.token_exchange_post_data_method: token_exchange_data}
        r = self.session.post(self.config.token_uri, **post_request_kwargs)
        return r.json()[self.config.token_key]

    def _constant_or_xpath_extracted(self, value, form_element):
        match = self.CONFIG_XPATH_REGEX.match(value)
        if not match:
            return value
        value_from_xpath = form_element.xpath(self.CONFIG_XPATH_REGEX.match(value).groupdict('xpath_value'))
        if len(value_from_xpath) == 1:
            return value_from_xpath[0]

    @staticmethod
    def compute_state():
        rand = SystemRandom()
        return ''.join(repeatfunc(rand.choice, 22, string.digits + string.ascii_lowercase + string.ascii_uppercase))


class CodeAuthorizedAuth(AuthBase):

    def __init__(self, token, where, key=None):
        self.token = token
        self.where = where
        self.key = key

    def __call__(self, request):
        if self.where == 'qs':
            parts = urlparse(request.url)
            qs = parse_qs(parts.query)
            qs[self.key] = self.token
            request.url = urlunparse((
                parts.scheme,
                parts.netloc,
                parts.path,
                parts.params,
                urlencode(qs),
                parts.fragment
            ))
        elif self.where == 'header':
            request.headers['Authorization'] = "Bearer {}".format(self.token)
        return request



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

import logging
import re

from eodag.plugins.crunch.base import Crunch
from eodag.utils.exceptions import MisconfiguredError


logger = logging.getLogger('eodag.plugins.crunch.filter_latest')


class FilterLatestByName(Crunch):
    NAME_PATTERN_CONSTRAINT = re.compile(r'\(\?P<tileid>\\d\{6\}\)')

    def __init__(self, config):
        super(FilterLatestByName, self).__init__(config)
        name_pattern = config.pop('name_pattern')
        if not self.NAME_PATTERN_CONSTRAINT.search(name_pattern):
            raise MisconfiguredError('Name pattern should respect the regex: {}'.format(
                self.NAME_PATTERN_CONSTRAINT.pattern
            ))
        self.name_pattern = re.compile(name_pattern)

    def proceed(self, product_list, **search_params):
        """Filter Search results to get only the latest product, based on the name of the product"""
        logger.debug('Starting products filtering')
        processed = []
        filtered = []
        for product in product_list:
            match = self.name_pattern.match(product.properties['title'])
            if match:
                tileid = match.group('tileid')
                if tileid not in processed:
                    logger.debug('Latest product found for tileid=%s: date=%s', tileid,
                                 product.properties['startTimeFromAscendingNode'])
                    filtered.append(product)
                    processed.append(tileid)
                else:
                    logger.debug('Latest product already found for tileid=%s', tileid)
            else:
                logger.warning('The name of the product %r as returned by the search plugin does not match the name '
                               'pattern expected by the cruncher %s. Name of the product: %s. Name pattern expected: '
                               '%s', product, self.name, product.properties['title'], self.name_pattern)
        logger.debug('Ending products filtering. Filtered products: %r', filtered)
        return filtered



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .api.core import EODataAccessGateway      # noqa

# Filter Cython warnings
# Apparently (see
# https://stackoverflow.com/questions/40845304/runtimewarning-numpy-dtype-size-changed-may-indicate-binary-incompatibility     # noqa
# and https://github.com/numpy/numpy/issues/11788) this is caused by Cython and affect pre-built Python packages that
# depends on numpy and ship with a pre-built version of numpy that is older than 1.15.1 (where the warning is silenced
# exactly as below)
import warnings
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import logging.config


def setup_logging(**kwargs):
    verbosity = kwargs.pop('verbose')
    if verbosity == 0:
        logging.config.dictConfig({
            'version': 1,
            'disable_existing_loggers': False,
            'handlers': {
                'null': {
                    'level': 'DEBUG',
                    'class': 'logging.NullHandler',
                },
            },
            'loggers': {
                'eodag': {
                    'handlers': ['null'],
                    'propagate': True,
                    'level': 'INFO',
                },
            }
        })
    elif verbosity <= 2:
        logging.config.dictConfig({
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'standard': {
                    'format': '%(asctime)s-15s %(name)-32s [%(levelname)-8s] %(message)s'
                },
            },
            'handlers': {
                'console': {
                    'level': 'DEBUG',
                    'class': 'logging.StreamHandler',
                    'formatter': 'standard'
                },
            },
            'loggers': {
                'eodag': {
                    'handlers': ['console'],
                    'propagate': True,
                    'level': 'INFO',
                },
                'sentinelsat': {
                    'handlers': ['console'],
                    'propagate': True,
                    'level': 'INFO',
                },
            }
        })
    elif verbosity == 3:
        logging.config.dictConfig({
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'verbose': {
                    'format': '%(asctime)s-15s %(name)-32s [%(levelname)-8s] (%(module)-17s) %(message)s'
                },
            },
            'handlers': {
                'console': {
                    'level': 'DEBUG',
                    'class': 'logging.StreamHandler',
                    'formatter': 'verbose'
                },
            },
            'loggers': {
                'eodag': {
                    'handlers': ['console'],
                    'propagate': True,
                    'level': 'DEBUG',
                },
                'sentinelsat': {
                    'handlers': ['console'],
                    'propagate': True,
                    'level': 'DEBUG',
                },
            }
        })



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import logging
import os

import boto3

from eodag.plugins.download.base import Download
from eodag.utils import urlparse

logger = logging.getLogger('eodag.plugins.download.aws')


class AwsDownload(Download):

    def download(self, product, auth=None, progress_callback=None):
        bucket_name, prefix = self.get_bucket_name_and_prefix(product)
        access_key, access_secret = auth
        s3 = boto3.resource('s3', aws_access_key_id=access_key, aws_secret_access_key=access_secret)
        bucket = s3.Bucket(bucket_name)

        for product_chunk in bucket.objects.filter(Prefix=prefix):
            product_local_path = os.path.join(
                self.config.outputs_prefix,
                product.properties.get('title', product.properties['id'])
            )
            if not os.path.isdir(product_local_path):
                os.makedirs(product_local_path)

            chunck_path_as_list = product_chunk.key.split('/')
            if len(chunck_path_as_list) > 9:
                chunk_dir = os.path.join(product_local_path, chunck_path_as_list[-2])
                if not os.path.isdir(chunk_dir):
                    os.makedirs(chunk_dir)
                chunk_file_path = os.path.join(chunk_dir, chunck_path_as_list[-1])
            else:
                chunk_file_path = os.path.join(product_local_path, chunck_path_as_list[-1])
            if not os.path.isfile(chunk_file_path):
                bucket.download_file(product_chunk.key, chunk_file_path)
            return product_local_path

    def get_bucket_name_and_prefix(self, product):
        bucket, prefix = None, None
        # Assume the bucket is encoded into the product location as a URL or given as the 'associated_bucket' config
        # param
        scheme, netloc, path, params, query, fragment = urlparse(product.location)
        if not netloc and scheme == 's3':
            bucket, path = self.config.associated_bucket, path
        elif scheme in ('http', 'https', 'ftp'):
            parts = path.split('/')
            bucket, prefix = parts[1], '/{}'.format('/'.join(parts[2:]))
        return bucket, prefix



# -*- coding: utf-8 -*-
# Copyright 2018, CS Systemes d'Information, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

import re
from datetime import datetime
from string import Formatter

from dateutil.tz import tzutc
from lxml import etree
from lxml.etree import XPathEvalError
from shapely import geometry
from six import string_types

from eodag.utils import get_timestamp

SEP = r"#"
INGEST_CONVERSION_REGEX = re.compile(
    r"^{(?P<path>[^#]*)" + SEP + r"(?P<converter>[^\d\W]\w*)}$"
)
NOT_AVAILABLE = "Not Available"
NOT_MAPPED = "Not Mapped"


def get_metadata_path(map_value):
    """Return the jsonpath or xpath to the value of a EO product metadata in a provider
    search result.

    The path is retrieved depending on if the metadata is queryable (the value
    associated to it in the provider search config metadata mapping is a list) or not
    (the value is directly the string corresponding to the path).

    Assume we have the following provider config::

        provider:
            ...
            search:
                ...
                metadata_mapping:
                    productType:
                        - productType
                        - $.properties.productType
                    id: $.properties.id
                    ...
                ...
            ...

    Then the metadata `id` is not queryable for this provider meanwhile `productType`
    is queryable. The first value of the `metadata_mapping.productType` is how the
    eodag search parameter `productType` is interpreted in the
    :class:`~eodag.plugins.search.base.Search` plugin implemented by `provider`, and is
    used when eodag delegates search process to the corresponding plugin.

    :param map_value: The value originating from the definition of `metadata_mapping`
                      in the provider search config. For example, it is the list
                      `['productType', '$.properties.productType']` with the sample
                      above
    :type map_value: str (Python 3) or unicode (Python 2) or list(str or unicode)
    :return: The value of the path to the metadata value in the provider search result
    :rtype: str (Python 3) or unicode (Python 2)
    """
    path = map_value[1] if isinstance(map_value, list) else map_value
    match = INGEST_CONVERSION_REGEX.match(path)
    if match:
        g = match.groupdict()
        return g["converter"], g["path"]
    return None, path


def get_search_param(map_value):
    """See :func:`~eodag.api.product.metadata_mapping.get_metadata_path`

    :param map_value: The value originating from the definition of `metadata_mapping`
                      in the provider search config
    :type map_value: list
    :return: The value of the search parameter as defined in the provider config
    :rtype: str (Python 3) or unicode (Python 2)
    """
    # Assume that caller will pass in the value as a list
    return map_value[0]


def format_metadata(search_param, *args, **kwargs):
    """Format a string of form {<field_name>#<conversion_function>}

    The currently understood converters are:
        - ``to_timestamp_milliseconds``: converts a utc date string to a timestamp in
          milliseconds
        - ``to_wkt``: converts a geometry to its well known text representation
        - ``to_iso_utc_datetime_from_milliseconds``: converts a utc timestamp in given
          milliseconds to a utc iso datetime
        - ``to_iso_utc_datetime``: converts a UTC datetime string to ISO UTC datetime
          string
        - ``to_iso_date``: removes the time part of a iso datetime string
        - ``remove_extension``: on a string that contains dots, only take the first
          part of the list obtained by splitting the string on dots

    :param search_param: The string to be formatted
    :type search_param: str or unicode
    :param args: (optional) Additional arguments to use in the formatting process
    :type args: tuple
    :param kwargs: (optional) Additional named-arguments to use when formatting
    :type kwargs: dict
    :returns: The formatted string
    :rtype: str or unicode

    .. versionadded::
        1.0

            * Added the ``remove_extension`` metadata converter
    """

    class MetadataFormatter(Formatter):
        CONVERSION_REGEX = re.compile(
            r"^(?P<field_name>.+)" + SEP + r"(?P<converter>[^\d\W]\w*)$"
        )

        def __init__(self):
            self.custom_converter = None

        def get_field(self, field_name, args, kwargs):
            conversion_func_spec = self.CONVERSION_REGEX.match(field_name)
            # Register a custom converter if any for later use (see convert_field)
            # This is done because we don't have the value associated to field_name at
            # this stage
            if conversion_func_spec:
                field_name = conversion_func_spec.groupdict()["field_name"]
                converter = conversion_func_spec.groupdict()["converter"]
                self.custom_converter = getattr(self, "convert_{}".format(converter))
            return super(MetadataFormatter, self).get_field(field_name, args, kwargs)

        def convert_field(self, value, conversion):
            # Do custom conversion if any (see get_field)
            if self.custom_converter is not None:
                converted = self.custom_converter(value) if value is not None else ""
                # Clear this state variable in case the same converter is used to
                # resolve other named arguments
                self.custom_converter = None
                return converted
            return super(MetadataFormatter, self).convert_field(value, conversion)

        @staticmethod
        def convert_to_timestamp_milliseconds(value):
            return int(1e3 * get_timestamp(value))

        @staticmethod
        def convert_to_wkt(value):
            return geometry.box(
                *[
                    float(v)
                    for v in "{lonmin} {latmin} {lonmax} {latmax}".format(
                        **value
                    ).split()
                ]
            ).to_wkt()

        @staticmethod
        def convert_to_iso_utc_datetime_from_milliseconds(timestamp):
            try:
                return datetime.fromtimestamp(timestamp / 1e3, tzutc()).isoformat()
            except TypeError:
                return timestamp

        @staticmethod
        def convert_to_iso_utc_datetime(dt):
            for idx, fmt in enumerate(("%Y-%m-%d", "%Y-%m-%dT%H:%M:%S")):
                try:
                    return (
                        datetime.strptime(dt, fmt)
                        .replace(tzinfo=tzutc())
                        .isoformat()
                        .replace("+00:00", "Z")
                    )
                except ValueError:
                    if idx == 1:
                        raise

        @staticmethod
        def convert_to_iso_date(datetime_string):
            return datetime_string[:10]

        @staticmethod
        def convert_remove_extension(string):
            parts = string.split(".")
            if parts:
                return parts[0]
            return ""

    return MetadataFormatter().vformat(search_param, args, kwargs)


def properties_from_json(json, mapping):
    """Extract properties from a provider json result.

    :param json: the representation of a provider result as a json object
    :type json: dict
    :param mapping: a mapping between :class:`~eodag.api.product.EOProduct`'s metadata
                    keys and the location of the values of these properties in the json
                    representation, expressed as a
                    `jsonpath <http://goessner.net/articles/JsonPath/>`_
    :return: the metadata of the :class:`~eodag.api.product.EOProduct`
    :rtype: dict
    """
    properties = {}
    templates = {}
    for metadata, value in mapping.items():
        # Treat the case when the value is from a queryable metadata
        if isinstance(value, list):
            conversion_or_none, path_or_text = value[1]
        else:
            conversion_or_none, path_or_text = value
        if isinstance(path_or_text, string_types):
            if re.search(r"({[^{}]+})+", path_or_text):
                templates[metadata] = path_or_text
            else:
                properties[metadata] = path_or_text
        else:
            match = path_or_text.find(json)
            extracted_value = match[0].value if len(match) == 1 else NOT_AVAILABLE
            if extracted_value is None:
                properties[metadata] = None
            else:
                if conversion_or_none is None:
                    properties[metadata] = extracted_value
                else:
                    properties[metadata] = format_metadata(
                        "{%s%s%s}" % (metadata, SEP, conversion_or_none),
                        **{metadata: extracted_value}
                    )
    # Resolve templates
    for metadata, template in templates.items():
        properties[metadata] = template.format(**properties)
    return properties


def properties_from_xml(xml_as_text, mapping, empty_ns_prefix="ns"):
    """Extract properties from a provider xml result.

    :param xml_as_text: the representation of a provider result as xml
    :type xml_as_text: str or unicode
    :param mapping: a mapping between :class:`~eodag.api.product.EOProduct`'s metadata
                    keys and the location of the values of these properties in the xml
                    representation, expressed as a
                    `xpath <https://www.w3schools.com/xml/xml_xpath.asp>`_
    :param empty_ns_prefix: the name to give to the default namespace of `xml_as_text`.
                            This is a technical workaround for the limitation of lxml
                            not supporting empty namespace prefix (default: ns). The
                            xpath in `mapping` must use this value to be able to
                            correctly reach empty-namespace prefixed elements
    :type empty_ns_prefix: str or unicode
    :return: the metadata of the :class:`~eodag.api.product.EOProduct`
    :rtype: dict
    """
    properties = {}
    templates = {}
    root = etree.XML(xml_as_text)
    for metadata, value in mapping.items():
        # Treat the case when the value is from a queryable metadata
        if isinstance(value, list):
            conversion_or_none, path_or_text = value[1]
        else:
            conversion_or_none, path_or_text = value
        try:
            extracted_value = root.xpath(
                path_or_text,
                namespaces={k or empty_ns_prefix: v for k, v in root.nsmap.items()},
            )
            if len(extracted_value) == 1:
                if conversion_or_none is None:
                    properties[metadata] = extracted_value[0]
                else:
                    properties[metadata] = format_metadata(
                        "{%s%s%s}" % (metadata, SEP, conversion_or_none),
                        **{metadata: extracted_value[0]}
                    )
            # If there are multiple matches, consider the result as a list, doing a
            # formatting if any
            elif len(extracted_value) > 1:
                if conversion_or_none is None:
                    properties[metadata] = extracted_value
                else:
                    properties[metadata] = [
                        format_metadata(
                            "{%s%s%s}"
                            % (
                                metadata,
                                SEP,
                                conversion_or_none,
                            ),  # Re-build conversion format identifier
                            **{metadata: extracted_value_item}
                        )
                        for extracted_value_item in extracted_value
                    ]
            # If there is no matched value (empty list), mark the metadata as not
            # available
            else:
                properties[metadata] = NOT_AVAILABLE
        except XPathEvalError:
            # Assume the mapping is to be passed as is, in which case we readily
            # register it, or is a template, in which case we register it for later
            # formatting resolution using previously successfully resolved properties
            # Ignore any transformation specified. If a value is to be passed as is,
            # we don't want to transform it further
            if re.search(r"({[^{}]+})+", path_or_text):
                templates[metadata] = path_or_text
            else:
                properties[metadata] = path_or_text
    # Resolve templates
    for metadata, template in templates.items():
        properties[metadata] = template.format(**properties)
    return properties


# Keys taken from http://docs.opengeospatial.org/is/13-026r8/13-026r8.html
# For a metadata to be queryable, The way to query it must be specified in the
# provider metadata_mapping configuration parameter. It will be automatically
# detected as queryable by eodag when this is done
DEFAULT_METADATA_MAPPING = {
    # Opensearch resource identifier within the search engine context (in our case
    # within the context of the data provider)
    "uid": "$.uid",
    # OpenSearch Parameters for Collection Search (Table 3)
    "productType": "$.properties.productType",
    "doi": "$.properties.doi",
    "platform": "$.properties.platform",
    "platformSerialIdentifier": "$.properties.platformSerialIdentifier",
    "instrument": "$.properties.instrument",
    "sensorType": "$.properties.sensorType",
    "compositeType": "$.properties.compositeType",
    "processingLevel": "$.properties.processingLevel",
    "orbitType": "$.properties.orbitType",
    "spectralRange": "$.properties.spectralRange",
    "wavelengths": "$.properties.wavelengths",
    "hasSecurityConstraints": "$.properties.hasSecurityConstraints",
    "dissemination": "$.properties.dissemination",
    # INSPIRE obligated OpenSearch Parameters for Collection Search (Table 4)
    "title": "$.properties.title",
    "topicCategory": "$.properties.topicCategory",
    "keyword": "$.properties.keyword",
    "abstract": "$.properties.abstract",
    "resolution": "$.properties.resolution",
    "organisationName": "$.properties.organisationName",
    "organisationRole": "$.properties.organisationRole",
    "publicationDate": "$.properties.publicationDate",
    "lineage": "$.properties.lineage",
    "useLimitation": "$.properties.useLimitation",
    "accessConstraint": "$.properties.accessConstraint",
    "otherConstraint": "$.properties.otherConstraint",
    "classification": "$.properties.classification",
    "language": "$.properties.language",
    "specification": "$.properties.specification",
    # OpenSearch Parameters for Product Search (Table 5)
    "parentIdentifier": "$.properties.parentIdentifier",
    "productionStatus": "$.properties.productionStatus",
    "acquisitionType": "$.properties.acquisitionType",
    "orbitNumber": "$.properties.orbitNumber",
    "orbitDirection": "$.properties.orbitDirection",
    "track": "$.properties.track",
    "frame": "$.properties.frame",
    "swathIdentifier": "$.properties.swathIdentifier",
    "cloudCover": "$.properties.cloudCover",
    "snowCover": "$.properties.snowCover",
    "lowestLocation": "$.properties.lowestLocation",
    "highestLocation": "$.properties.highestLocation",
    "productVersion": "$.properties.productVersion",
    "productQualityStatus": "$.properties.productQualityStatus",
    "productQualityDegradationTag": "$.properties.productQualityDegradationTag",
    "processorName": "$.properties.processorName",
    "processingCenter": "$.properties.processingCenter",
    "creationDate": "$.properties.creationDate",
    "modificationDate": "$.properties.modificationDate",
    "processingDate": "$.properties.processingDate",
    "sensorMode": "$.properties.sensorMode",
    "archivingCenter": "$.properties.archivingCenter",
    "processingMode": "$.properties.processingMode",
    # OpenSearch Parameters for Acquistion Parameters Search (Table 6)
    "availabilityTime": "$.properties.availabilityTime",
    "acquisitionStation": "$.properties.acquisitionStation",
    "acquisitionSubType": "$.properties.acquisitionSubType",
    "startTimeFromAscendingNode": "$.properties.startTimeFromAscendingNode",
    "completionTimeFromAscendingNode": "$.properties.completionTimeFromAscendingNode",
    "illuminationAzimuthAngle": "$.properties.illuminationAzimuthAngle",
    "illuminationZenithAngle": "$.properties.illuminationZenithAngle",
    "illuminationElevationAngle": "$.properties.illuminationElevationAngle",
    "polarizationMode": "$.properties.polarizationMode",
    "polarisationChannels": "$.properties.polarisationChannels",
    "antennaLookDirection": "$.properties.antennaLookDirection",
    "minimumIncidenceAngle": "$.properties.minimumIncidenceAngle",
    "maximumIncidenceAngle": "$.properties.maximumIncidenceAngle",
    "dopplerFrequency": "$.properties.dopplerFrequency",
    "incidenceAngleVariation": "$.properties.incidenceAngleVariation",
    # Custom parameters (not defined in the base document referenced above)
    # id differs from uid. The id is an identifier by which a product which is
    # distributed by many providers can be retrieved (a property that it has in common
    # in the catalogues of all the providers on which it is referenced)
    "id": "$.id",
    # The geographic extent of the product
    "geometry": "$.geometry",
    # The url of the quicklook
    "quicklook": "$.properties.quicklook",
    # The url to download the product "as is" (literal or as a template to be completed
    # either after the search result is obtained from the provider or during the eodag
    # download phase)
    "downloadLink": "$.properties.downloadLink",
    # custom field, can contain several parameters : "key1=value1&key2=value2"
    "custom": "$.properties.custom",
}
